import os

import numpy as np

import jax
import jax.numpy as jnp
# As of jax v0.4.24, jax.numpy.trapz has been deprecated, and replaced
# by scipy.integrate.trapezoid.
try: 
    import jax.numpy.trapz as trapz 
except ImportError: 
    from jax.scipy.integrate import trapezoid as trapz

import equinox as eqx

from . import const as const 
from .special_funcs import gamma 
from jax.scipy.special import spence
from jax.scipy.special import expit

file_dir = os.path.dirname(__file__)

# Particle masses
from .const import me, mn, mp
Q = mn - mp # Mass difference between neutrons and protons

class WeakRates(eqx.Module): 
    """
    Class for weak rates calculation. 

    Attributes
    ----------
    RC_corr : bool
        Radiative corrections to weak rates. 
    thermal_corr : bool
        Thermal and bremsstrahlung corrections to weak rates 
        (pre-tabulated, assuming standard BBN). 
    FM_corr : bool
        Finite mass corrections to weak rates. 
    weak_mag_corr : bool
        Weak magnetism corrections to weak rates. 
    T_nTOp_thermal_interval : array
        EM temperature abscissa for tabulated n->p thermal corrections. 
    T_pTOn_thermal_interval : array
        EM temperature abscissa for tabulated p->n thermal corrections.
    L_nTOpCCRTh_res : array
        Tabulated dimensionless thermal corrections to n->p rate. 
        Divide by (tau_n * lambda_0) for the actual correction. 
    L_pTOnCCRTh_res : array
        Tabulated dimensionless thermal corrections to p->n rate. 
        Divide by (tau_n * lambda_0) for the actual correction. 
    """

    RC_corr : bool
    thermal_corr : bool
    FM_corr : bool
    weak_mag_corr : bool

    T_nTOp_thermal_interval : list
    T_pTOn_thermal_interval : list
    L_nTOpCCRTh_res : list
    L_pTOnCCRTh_res : list 

    lambda_0 : float

    def __init__(self, 
        RC_corr=True, FM_corr=True, weak_mag_corr=True, 
        thermal_corr=True
    ): 
        """
        Parameters
        ----------
        RC_corr : bool
            Include radiative corrections. 
        FM_corr : bool
            Include finite mass corrections. 
        weak_mag_corr : bool
            Include weak magnetism corrections. 
        thermal_corr : bool
            Include thermal and bremsstrahlung corrections. 
        """
        
        self.RC_corr = RC_corr
        self.FM_corr = FM_corr 
        self.weak_mag_corr = weak_mag_corr
        self.thermal_corr = thermal_corr
        
        if self.thermal_corr: 

            self.T_nTOp_thermal_interval, self.L_nTOpCCRTh_res = np.loadtxt(
                file_dir+"/data/weak_thermal_corrections/"
                +"nTOp_thermal_corrections_SBBN.txt", 
                unpack = True
            )

            self.T_pTOn_thermal_interval, self.L_pTOnCCRTh_res = np.loadtxt(
                file_dir+"/data/weak_thermal_corrections/"
                +"pTOn_thermal_corrections_SBBN.txt", 
                unpack = True
            )

        else: 

            self.T_nTOp_thermal_interval = [] 
            self.T_pTOn_thermal_interval = [] 
            self.L_nTOpCCRTh_res = [] 
            self.L_pTOnCCRTh_res = [] 

        self.lambda_0 = 0. 

        # Slight shift in limits to avoid unimportant singularities.
        en_vals = jnp.linspace(1.+.1e-6, Q/me-1e-6, 1000)
        dlambda_den_vals = self.dlambda_den_RC(en_vals)
        self.lambda_0 += trapz(dlambda_den_vals, en_vals)

        if self.FM_corr:

            pe_vals = jnp.linspace(
                0.+1e-4, jnp.sqrt((Q/me)**2 - 1.)-1e-5, 1000
            )
            y_vals = self.dlambda_dp_FM(pe_vals)
            self.lambda_0 += trapz(y_vals, pe_vals)

    @eqx.filter_jit
    def __call__(
        self, T_vec_ref, T_start, T_end, sampling_nTOp
    ): 
        """
        Evaluate n <-> p rates over range of EM temperatures. 

        Parameters
        ----------
        T_vec_ref : tuple of ndarray
            Reference (T_gamma, T_nu) to evaluate weak rates. In MeV. 
        T_start : float
            Highest photon temperature to evaluate rates at. In MeV.   
        T_end : float
            Lowest photon temperature to evaluate rates at. In MeV. 
        sampling_nTOp : int
            Number of points between T_start and T_end to evaluate at. 

        Returns
        -------
        tuple
            (T_EM abscissa, n->p rates, p->n rates). Note that rates are 
            dimensionless, normalized to the neutron decay width. T_EM abscissa
            limits are T_start, T_end, binned according to sampling_nTOp. 
        
        
        """
    
        T_interval = jnp.logspace(
            jnp.log10(T_start), jnp.log10(T_end), sampling_nTOp
        )

        nTOp_rates = self.nTOp_rates(T_interval, T_vec_ref)

        return (T_interval, ) + nTOp_rates

    @eqx.filter_vmap(in_axes=(None, 0, None))
    def nTOp_rates(self, Tg, T_vec_ref): 
        """
        Dimensionless n <-> p rates, normalized to neutron decay width.

        Parameters
        ----------
        Tg : float
            EM temperature in MeV at which to evaluate rate. 
        T_vec_ref : tuple
            (EM temperature array, neutrino temperature array) in MeV, used 
            for computing the weak rates. 

        Returns
        -------
        tuple
            (n -> p rate, p -> n rate). Note that rates are 
            dimensionless, normalized to the neutron decay width. 
        """

        Tg_vec_ref, Tnu_vec_ref = T_vec_ref
        Tnu_of_Tg_ref = Tnu_vec_ref / Tg_vec_ref
        
        x = me / Tg
        xnu = me / (
            Tg * jnp.interp(
                Tg, jnp.flip(Tg_vec_ref), jnp.flip(Tnu_of_Tg_ref), left=Tnu_of_Tg_ref[-1], right=Tnu_of_Tg_ref[0]
            )
        )

        pemin = 0.00001 
        
        pemax = jnp.maximum(7., 30./x) 
        p_vals = jnp.linspace(pemin+1e-4, pemax-1e-5, 1000) 

        nTOp_rate = 0. 
        pTOn_rate = 0. 

        y_CCR_vals = jnp.array([
            self.dGamma_nTOp_dp(p_vals, x, xnu), 
            self.dGamma_pTOn_dp(p_vals, x, xnu)
        ])
        CCR_rates = trapz(y_CCR_vals, p_vals)
        nTOp_rate += CCR_rates[0] / self.lambda_0 
        pTOn_rate += CCR_rates[1] / self.lambda_0 

        if self.FM_corr:
    
            y_FMCCR_vals = jnp.array([ 
                self.ddelt_Gamma_nTOp_FM_dp(p_vals, x, xnu),
                self.ddelt_Gamma_pTOn_FM_dp(p_vals, x, xnu)
            ])
            FMCCR_rates = trapz(y_FMCCR_vals, p_vals)
            nTOp_rate += FMCCR_rates[0] / self.lambda_0 
            pTOn_rate += FMCCR_rates[1] / self.lambda_0 

        if self.thermal_corr: 

            thermal_rates = jnp.array([
                jnp.interp(
                    Tg, 
                    const.kB * self.T_nTOp_thermal_interval, 
                    self.L_nTOpCCRTh_res,
                    left=self.L_nTOpCCRTh_res[0],
                    right=self.L_nTOpCCRTh_res[-1]
                ),
                jnp.interp(
                    Tg, 
                    const.kB * self.T_pTOn_thermal_interval, 
                    self.L_pTOnCCRTh_res,
                    left=self.L_pTOnCCRTh_res[0], 
                    right=self.L_pTOnCCRTh_res[-1]
                )
            ])
            nTOp_rate += thermal_rates[0] / self.lambda_0 
            pTOn_rate += thermal_rates[1] / self.lambda_0 

        return (nTOp_rate, pTOn_rate)

    
    def Sirlin_G(self, kmax, en):
        """
        Sirlin's universal function. 

        Parameters
        ----------
        kmax : float
            Dimensionless maximum energy of photon considered in 
            radiative correction, normalized to electron mass. 
        en : float
            Dimensionless electron energy, normalized to electron mass)

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (B32). 

        """

        b = jnp.sqrt(en**2 - 1.) / en

        Rd_b = jnp.where(b==0, 1, jnp.arctanh(b) / b)

        return (
            3.*jnp.log(mp/me) - 3./4. + 4.*(Rd_b - 1.) * (
                kmax/(3.*en) - 3./2. + jnp.log(2.*kmax)
            ) 
            + Rd_b * (2.*(1. + b**2) + kmax**2/(6.*en**2) - 4.*b*Rd_b) 
            + (4./b)*(-spence(1. - 2.*b/(1. + b)))
        )

    def R_RC(self, kmax, en):
        """
        Resummed radiative correction term at T = 0. 

        Parameters
        ----------
        kmax : float
            Dimensionless maximum energy of photon considered in 
            radiative correction, normalized to electron mass. 
        en : float
            Dimensionless electron energy, normalized to electron mass)

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (103), full definition in Eq. (B35).
        Note that the constant 1/134 below is approximately the fine-structure
        constant at the proton mass scale.

        """

        # Constants defined in Pitrou+ 1801.08023 Eq. (B30)

        C = 0.891
        A_g = -0.34
        mA = 1.2e3 # in MeV 

        # Constants defined in Pitrou+ 1801.08023 Eq. (B36)
        L = 1.02094
        S = 1.02248 
        delta_factor = -0.00043 # alpha_FS/(2*pi) * delta
        NLL = -1e-4
            
        return (
            (
                1. + const.aFS / (2.*jnp.pi) * (
                    self.Sirlin_G(kmax, en) - 3.*jnp.log(mp / (2*Q))
                )
            ) * (L + (const.aFS/jnp.pi)*C + delta_factor) * (
                S + 1./(134.*2.*jnp.pi)*(jnp.log(mp/mA) + A_g) + NLL
            )
        )

    def Fermi(self, b):
        """
        Fermi function in the relativistic limit. 

        Parameters
        ----------
        b : float
            Speed of the electron. 
        
        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (100). Equivalent to Sommerfeld 
        enhancement factor when v << 1. 
        """

        Gamma = jnp.sqrt(1. - const.aFS**2.) - 1.
        gamma1 = 1. + Gamma
        gamma2 = 3. + 2.*Gamma
        lambda_Compton = 1. / (me / (const.hbar * const.c))

        return (
            (1. + Gamma/2.) * 4. 
            * (2. * const.radproton * b / lambda_Compton)**(2.*Gamma)
            / gamma(gamma2)**2
            * jnp.exp(jnp.pi * const.aFS / b) / (1. - b**2)**Gamma 
            * jnp.abs(gamma(gamma1 + const.aFS/b*1j))**2
        )
        
    def bFermi(self, b):
        """
        Fermi function in the relativistic limit, times speed of electron. 

        Parameters
        ----------
        b : float
            Speed of the electron. 

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (100). Equivalent to Sommerfeld 
        enhancement factor when v << 1. 

        """

        return b*self.Fermi(b)
    
    @eqx.filter_vmap(in_axes=(None, 0))
    def dlambda_den_RC(self, en):
        """
        Derivative of lambda with respect to energy of electron, 
        including radiative corrections at T = 0. 

        Parameters
        ----------
        en : float
            Energy of the electron produced in neutron decay, normalized to 
            the electron mass. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (106), but note the change in variables. 

        """

        q = Q/me
        b = jnp.sqrt(en**2 - 1.)/en

        if self.RC_corr: 
            R_term = (
                self.Fermi(b)
                * self.R_RC(q-en, en)
            )
        else: 
            R_term = 1. 

        return (
            en**2 * (q-en)**2 * b * R_term
        )
    
    def chi_n_decay(self, pe):
        """
        Finite nucleon mass correction term for neutron decay. 

        Parameters
        ----------
        pe : float
            Dimensionless momentum of the electron, normalized to electron 
            mass. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (B26). 

        """

        en = jnp.sqrt(pe**2 + 1.)

        # Equivalent to 2*fWM in Pitrou+ 1801.08023 Eq (B9). 
        # See also Ivanov+ 1212.0332
        if self.weak_mag_corr:
            delta_kappa = 3.7058
        else: 
            delta_kappa = 0. 

        M = mn/me
        q = Q/me
        
        # Pitrou+ 1801.08023 Eq. (B24a)--(B24c)
        f1 = (
            ((1. + const.gA)**2. + 2.*delta_kappa*const.gA)
            / (1. + 3.*const.gA**2)
        )
        f2 = (
            ((1. - const.gA)**2. - 2.*delta_kappa*const.gA)
            / (1. + 3.*const.gA**2)
        )
        f3 = (const.gA**2 - 1.) / (1. + 3.*const.gA**2)

        return (
            f1*(q - en)**2*(pe**2/(M*en)) 
            + f2/M*(q - en)**3 
            - (f1 + f2 + f3) / M * (
                2*(q - en)**3 + (q - en)*pe**2
            ) 
            + f3 * (q - en)**2 * (pe**2) / (M*en)
        )
    
    @eqx.filter_vmap(in_axes=(None, 0))
    def dlambda_dp_FM(self, pe):
        """
        Derivative of the correction to lambda with respect to momentum, 
        due to finite mass effects. 

        Parameters
        ----------
        pe : float
            Dimensionless momentum of the electron, normalized to the electron  
            mass.

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (118). If radiative correction is included, 
        add a factor of R. Derivative with respect to momentum (and not energy) 
        to keep consistency with PRyMordial. 

        """

        en = jnp.sqrt(pe**2 + 1.)
        b = pe/en

        if self.RC_corr:

            R_rad = self.R_RC(Q/me - en, en) 

        else: 

            R_rad = 1. 

        return (
            pe**2 * self.chi_n_decay(pe) 
            * R_rad * self.Fermi(b)
        )

    def chi_Born(self, en, x, x_nu, sgnq):
        """
        Integrand in momentum integral for Born weak rate. 

        Parameters
        ----------
        en : float
            Dimensionless energy of the electron, normalized to electron mass.
        x : float
            Dimensionless inverse EM temperature, normalized to electron mass. 
        x_nu : float 
            Dimensionless inverse nu temperature, normalized to electron mass. 
        sqnq : int
            Should have value +1 or -1, to switch between chi_+ and chi_-. 

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (79). 

        """

        # xi_nu = 0. # nu chemical potential set to zero for now. 

        # sgnq = +1 corresponds to chi_plus. 
        e_nu = en - sgnq*(Q/me) 
        g_nu = expit(-x_nu*e_nu)
        g_e  = expit(-x*(-en))

        return e_nu**2 * g_nu * g_e 
    
    def Fermi_sgn(self, sgnq, sgnE, b):
        """
        Fermi function, with a check for proton and electron final state. 

        Parameters
        ----------
        sgnq : int
            +1 or -1 to choose between F_+ and F_-. 
        sgnE : int
            +1 or -1 to choose between positive or negative arguments of F. 
        b : float
            The speed of the electron. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (102). 
        """
        
        result = jnp.where(sgnq*sgnE > 0, self.Fermi(b), 1.)
        return result
    
    ###############################
    # n<->p Rates                 #
    ###############################
       
    def dGamma_dp(self, p, x, x_nu, sgnq):
        """
        Integrand over momentum for n <-> p rate.  including radiative 
        corrections.

        Parameters
        ----------
        p : float
            Dimensionless momentum of the electron, normalized to the 
            electron mass. 
        x : float
            Dimensionless EM inverse temperature, normalized to the electron 
            mass. 
        x_nu : float
            Dimensionless neutrino inverse temperature, normalized to the
            electron mass. 
        sgnq : int
            +1 or -1, to select between n -> p or p -> n. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (101) and (104). Radiative corrections 
        included as desired. Without radiative corrections, this gives the 
        Born approximation result. 
        """

        en = jnp.sqrt(p**2 + 1.)

        if self.RC_corr: 
            RC_term_plus = (
                self.R_RC(jnp.abs(sgnq*Q/me - en), en)
                * self.Fermi_sgn(sgnq, 1, p/en)
            )
            RC_term_minus = (
                self.R_RC(jnp.abs(sgnq*Q/me + en), en)
                * self.Fermi_sgn(sgnq, -1, p/en)
            )
        else: 
            RC_term_plus = 1. 
            RC_term_minus = 1. 

        return p**2 * (
            (
                self.chi_Born(en, x, x_nu, sgnq)
                * RC_term_plus
            ) + (
                self.chi_Born(-en, x, x_nu, sgnq)
                * RC_term_minus
            )
        )

    @eqx.filter_vmap(in_axes=(None, 0, None, None))
    def dGamma_nTOp_dp(self, p, x, xnu):
        """
        Integrand over momentum for n -> p rate including radiative 
        corrections.

        Parameters
        ----------
        p : float
            Dimensionless momentum of the electron, normalized to the 
            electron mass. 
        x : float
            Dimensionless EM inverse temperature, normalized to the electron 
            mass. 
        x_nu : float
            Dimensionless neutrino inverse temperature, normalized to the
            electron mass. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (101). 
        """

        return self.dGamma_dp(p, x, xnu, 1)
    
    @eqx.filter_vmap(in_axes=(None, 0, None, None))
    def dGamma_pTOn_dp(self, p, x, xnu):
        """
        Integrand over momentum for p -> n rate including radiative 
        corrections.

        Parameters
        ----------
        p : float
            Dimensionless momentum of the electron, normalized to the 
            electron mass. 
        x : float
            Dimensionless EM inverse temperature, normalized to the electron 
            mass. 
        x_nu : float
            Dimensionless neutrino inverse temperature, normalized to the
            electron mass. 
        sgnq : int
            +1 or -1, to select between n -> p or p -> n. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (104). 
        """

        return self.dGamma_dp(p, x, xnu, -1)
    
    
    ###########################
    # Finite Mass Corrections #
    ###########################
       
       
    def chi_FM(self, en, x, x_nu, sgnq):
        """
        Integrand over momentum for finite mass correction to n <-> p rate.

        Parameters
        ----------
        en : float
            Dimensionless energy of the electron, normalized to the electron 
            mass. 
        x : float
            Dimensionless EM inverse temperature, normalized to the electron 
            mass. 
        x_nu : float
            Dimensionless neutrino inverse temperature, normalized to the 
            electron mass. 
        sgnq : int
            +1 or -1 corresponding to chi_+ or chi_-. 

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (B23). However, instead of m_N, this appears
        to choose m_p for + and m_n for -. This is consistent with PRIMAT, and 
        probably makes sense, since we should apply the finite mass correction 
        to the outgoing state. 

        Note also a typo on the first term Eq. (B23) (on dimensional grounds), 
        corrected in PRIMAT. It should be g_nu^(2,0) as well. 

        """
        pe = jnp.sqrt(en**2 - 1)

        Mp = mp/me
        Mn = mn/me
        
        # Toggles between proton and neutron mass, normalized to electron mass. 
        M_sgnq = (mp + mn - sgnq*Q)/(2. * me)

        # Equivalent to 2*fWM in Pitrou+ 1801.08023 Eq (B9). 
        # See also Ivanov+ 1212.0332
        if self.weak_mag_corr:
            delta_kappa = 3.7058
        else: 
            delta_kappa = 0. 

        f_1 = (
            (1. + sgnq*const.gA)**2. + 2.*delta_kappa*sgnq*const.gA
        ) / (1. + 3.*const.gA**2)
        f_2 = (
            (1. - sgnq*const.gA)**2. - 2.*delta_kappa*sgnq*const.gA
        )/(1. + 3.*const.gA**2)
        f_3 = (const.gA**2 - 1.)/(1. + 3.*const.gA**2)

        # Electron distribution function term. 
        g_e = expit(-x*(-en))

        # Dimensionless energy of the neutrino. 
        en_nu = en - sgnq*Q / me

        expit_neg = expit(-en_nu*x_nu)
        expit_pos = expit(en_nu*x_nu)

        res_e2p1 = (
            2 * en_nu * expit_neg**2 
            + (en_nu * expit_neg) * (2 - en_nu*x_nu) * expit_pos
        )
        
        res_e3p1 = (
            3 * en_nu**2 * expit_neg**2 
            + (en_nu**2 * expit_neg) * (3 - en_nu*x_nu) * expit_pos
        )
        
        res_e4p1 = (
            4 * en_nu**3 * expit_neg**2 
            + (en_nu**3 * expit_neg) * (4 - en_nu*x_nu) * expit_pos
        )
        
        res_e2p0 = en_nu**2 * expit_neg
        
        res_e3p0 = en_nu**3 * expit_neg
        
        res_e2p2 = (
            (en_nu*x_nu * (en_nu*x_nu - 4) + 2) * expit_neg * expit_pos**2 
            + (4 - en_nu*x_nu*(en_nu*x_nu + 4))*expit_neg**2 * expit_pos 
            + 2 * expit_neg**3
        )
        
        res_e3p2 = en_nu * (
            (en_nu*x_nu * (en_nu*x_nu - 6) + 6) * expit_neg * expit_pos**2 
            + (12 - en_nu*x_nu*(en_nu*x_nu + 6)) * expit_neg**2 * expit_pos 
            + 6 * expit_neg**3
        )
        
        res_e4p2 = en_nu**2 * (
            (en_nu*x_nu * (en_nu*x_nu - 8) + 12) * expit_neg * expit_pos**2 
            + (24 - en_nu*x_nu*(en_nu*x_nu + 8)) * expit_neg**2 * expit_pos 
            + 12 * expit_neg**3
        )

        result =  g_e * (
            f_1 * res_e2p0 * (pe**2 / (M_sgnq * en))
            + f_2 * res_e3p0 * (-(1. / M_sgnq))
            + (f_1+f_2+f_3) / (2. * x * M_sgnq) * (
                res_e4p2 + res_e2p2 * pe**2
            )
            + (f_1+f_2+f_3) / (2. * M_sgnq) * (
                res_e4p1 + res_e2p1 * pe**2
            )
            - (f_1+f_2) / (x * M_sgnq) * (
                res_e3p1 + res_e2p1 * pe**2 / (-en)
            )
            - f_3 * 3. / (x * M_sgnq) * res_e2p0 
            + f_3 / (3 * M_sgnq) * res_e3p1 * pe**2 / en
            + f_3 * 2. / (2. * x * 3. * M_sgnq) * res_e3p2 * pe**2 / en
            - (f_1 + f_2 + f_3) * 3. / (2. * x) * (
                (1. - (Mn/Mp)**sgnq) * res_e2p1
            )
        )
        return result

    def ddelt_Gamma_FM_dp(self, p, x, znu, sgnq):
        """
        Integrand over momentum for finite mass corrections to the n <-> p
        rate. 

        Parameters
        ----------
        p : float
            Dimensionless momentum of the electron, normalized to the 
            electron mass. 
        x : float
            Dimensionless EM inverse temperature, normalized to the electron 
            mass. 
        x_nu : float
            Dimensionless neutrino inverse temperature, normalized to the
            electron mass. 
        sgnq : int
            +1 or -1, to select between n -> p or p -> n. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (115). Radiative corrections included as 
        desired.
        """

        en = jnp.sqrt(p**2 + 1.)
        b = p / en

        if self.RC_corr: 
            RC_term_plus = self.R_RC(
                jnp.abs(sgnq*Q/me - en), en
            ) * self.Fermi_sgn(sgnq, 1, b) 
            RC_term_minus = self.R_RC(
                jnp.abs(sgnq*Q/me + en), en
            ) * self.Fermi_sgn(sgnq, -1, b)
        
        else: 
            RC_term_plus = 1. 
            RC_term_minus = 1. 

        result =  p**2 * (
            (
                self.chi_FM(en, x, znu, sgnq) 
                * RC_term_plus 
            ) + (
                self.chi_FM(-en, x, znu, sgnq) 
                * RC_term_minus
            )
        )
        return result

    @eqx.filter_vmap(in_axes=(None, 0, None, None))
    def ddelt_Gamma_nTOp_FM_dp(self, p, x, xnu):
        """
        Integrand over momentum for finite mass corrections to the n -> p
        rate. 

        Parameters
        ----------
        p : float
            Dimensionless momentum of the electron, normalized to the 
            electron mass. 
        x : float
            Dimensionless EM inverse temperature, normalized to the electron 
            mass. 
        x_nu : float
            Dimensionless neutrino inverse temperature, normalized to the
            electron mass. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (115). 
        """

        return self.ddelt_Gamma_FM_dp(p, x, xnu, 1)

    @eqx.filter_vmap(in_axes=(None, 0, None, None))
    def ddelt_Gamma_pTOn_FM_dp(self, p, x, xnu):
        """
        Integrand over momentum for finite mass corrections to the p -> n 
        rate. 

        Parameters
        ----------
        p : float
            Dimensionless momentum of the electron, normalized to the 
            electron mass. 
        x : float
            Dimensionless EM inverse temperature, normalized to the electron 
            mass. 
        x_nu : float
            Dimensionless neutrino inverse temperature, normalized to the
            electron mass. 

        Returns
        -------
        float

        Notes
        -----
        See Pitrou+ 1801.08023 Eq. (115). 
        """

        return self.ddelt_Gamma_FM_dp(p, x, xnu, -1)


import numpy as np

import jax.numpy as jnp
from jax import config, lax, grad
import equinox as eqx

from diffrax import diffeqsolve, SaveAt, ODETerm, Tsit5, Kvaerno3, PIDController, ForwardMode, Event

from abcmb import constants as cnst
from . import recomb_functions
from .array_with_padding import array_with_padding
config.update("jax_enable_x64", True)

import os
file_dir = os.path.dirname(__file__)

class hydrogen_model(eqx.Module):
    """
    Hydrogen recombination model implementation.

    Computes hydrogen ionization fraction evolution through multiple phases:
    Saha equilibrium, post-Saha expansion, HYREC-2 EMLA with two-photon processes,
    and late-time EMLA-only evolution.

    Methods:
    --------
    get_hydrogen_history : Compute full hydrogen recombination history (units: dimensionless)
    Saha_equilibrium : Compute Saha equilibrium phase (units: dimensionless)
    post_Saha_expansion : Compute post-Saha expansion phase (units: dimensionless)
    solve_emla_twophoton : Solve EMLA with two-photon processes (units: dimensionless)
    solve_emla : Solve EMLA-only evolution (units: dimensionless)
    xe_derivative_twophoton : Compute ionization fraction derivative with two-photon (units: dimensionless)
    xe_tm_derivative : Compute coupled xe and Tm derivatives (units: dimensionless, eV)
    dxe_dlna_twophoton : Compute two-photon recombination rate (units: dimensionless)
    get_current_correction_func : Interpolate SWIFT correction function (units: dimensionless)
    steady_state_equations : Set up steady-state level population equations (units: dimensionless)
    Lyn_esc_rate : Compute Lyman escape rate (units: s^{-1})
    """
    integration_spacing : jnp.float64
    swift : jnp.array
    concrete_axis_size : jnp.array

    xe_4He : jnp.array
    lna_4He : jnp.array

    last_4He_lna : jnp.float64
    twog_redshift : jnp.float64

    lna_end : jnp.float64

    adjoint : "diffrax.adjoint" = eqx.field(static=True)

    def __init__(self, xe_4He, lna_4He, lna_end, last_4He_lna, twog_redshift, integration_spacing = 5.0e-4, Nsteps=800, swift = jnp.array(np.loadtxt(file_dir+"/tabs/fit_swift.dat")), adjoint = ForwardMode):
        """
        Initialize hydrogen recombination model.

        Parameters:
        -----------
        xe_4He : array_with_padding
            Helium ionization fraction from previous calculation
        lna_4He : array_with_padding
            Log scale factor array from helium calculation
        lna_end : array
            Ending log scale factor
        last_4He_lna : float
            Final log scale factor from helium recombination
        integration_spacing : float, optional
            Step size for integration (default: 5.0e-4)
        Nsteps : int, optional
            Maximum number of integration steps (default: 800)
        swift : array, optional
            SWIFT correction function tabulation
        """
        self.integration_spacing = integration_spacing
        self.swift = swift

        # Define time axes
        self.lna_end = lna_end
        self.concrete_axis_size = jnp.zeros(Nsteps)

        # pull in helium
        self.xe_4He = xe_4He
        self.lna_4He = lna_4He

        self.last_4He_lna = last_4He_lna
        self.twog_redshift = twog_redshift
        self.adjoint = adjoint

    def __call__(self, args, rtol=1e-6, atol=1e-9,solver=Kvaerno3(),max_steps=1024):
        """
        Compute hydrogen recombination history.

        Parameters:
        -----------
        args : tuple
            Background cosmology and cosmological parameters (BG, params)
        rtol : float, optional
            Relative tolerance for ODE solver (default: 1e-6)
        atol : float, optional
            Absolute tolerance for ODE solver (default: 1e-9)
        solver : diffrax.Solver, optional
            ODE solver instance (default: Kvaerno3())
        max_steps : int, optional
            Maximum solver steps (default: 1024)

        Returns:
        --------
        tuple
            (xe_full, lna_full, Tm, lna_Tm) - ionization fraction, log scale factor,
            matter temperature, and temperature grid
        """
        return self.get_hydrogen_history(args, rtol, atol, solver, max_steps)
    
    def get_hydrogen_history(self, args, rtol=1e-6, atol=1e-9,solver=Kvaerno3(),max_steps=1024):
        """
        Compute complete hydrogen recombination history through all phases.

        Sequentially computes hydrogen ionization fraction through post-Saha
        expansion, HYREC-2 EMLA with two-photon processes, and late-time
        EMLA-only evolution phases.

        Parameters:
        -----------
        args : tuple
            Background cosmology and cosmological parameters (BG, params)
        rtol : float, optional
            Relative tolerance for ODE solver (default: 1e-6)
        atol : float, optional
            Absolute tolerance for ODE solver (default: 1e-9)
        solver : diffrax.Solver, optional
            ODE solver instance (default: Kvaerno3())
        max_steps : int, optional
            Maximum solver steps (default: 1024)

        Returns:
        --------
        tuple
            (xe_full, lna_full, Tm, lna_Tm) containing ionization fraction evolution,
            log scale factor grid, matter temperature, and temperature grid
        """

        # Start computing xe at different phases
        ################## move to H ################## 

        ### POST SAHA EXPANSION PHASE ###
        xe_output_post, lna_output_post = self.post_Saha_expansion(self.last_4He_lna+self.integration_spacing, args)

        xe_4He_and_post = self.xe_4He.concat(array_with_padding(xe_output_post))
        lna_4He_and_post = self.lna_4He.concat(array_with_padding(lna_output_post))

        ### END OF POST SAHA EXPANSION PHASE ###

        ### HYREC2 EMLA + FULL TWO PHOTON PHASE ###
        xe_output_2g, lna_output_2g = self.solve_emla_twophoton(lna_4He_and_post.lastval, -jnp.log(self.twog_redshift), xe_4He_and_post.lastval, args, rtol, atol, solver, max_steps)

        xe_4He_post_2g = xe_4He_and_post.concat(array_with_padding(xe_output_2g))
        lna_4He_post_2g = lna_4He_and_post.concat(array_with_padding(lna_output_2g))
        ### END HYREC2 EMLA + FULL TWO PHOTON PHASE ###

        ### HYREC2 EMLA ONLY PHASE ###
        xe_output_late, Tm_output_late, lna_output_late = self.solve_emla(lna_4He_post_2g.lastval, xe_4He_post_2g.lastval, args, rtol, atol, solver)

        lna_Tm = array_with_padding(lna_output_late)
        Tm = array_with_padding(Tm_output_late)

        xe_4He_post_2g_late = xe_4He_post_2g.concat(array_with_padding(xe_output_late))
        lna_4He_post_2g_late = lna_4He_post_2g.concat(lna_Tm)
        ### END OF HYREC2 EMLA ONLY PHASE ###

        ### Begin TLA phase ###
        xe_output_TLA, Tm_output_TLA, lna_output_TLA = self.solve_TLA(lna_Tm.lastval,
                                                                      xe_4He_post_2g_late.lastval, Tm.lastval, 
                                                                      args)

        xe_all = xe_4He_post_2g_late.concat(array_with_padding(xe_output_TLA))
        lna_all = lna_4He_post_2g_late.concat(array_with_padding(lna_output_TLA))
        Tm_all = Tm.concat(array_with_padding(Tm_output_TLA))
        lna_Tm_all = lna_Tm.concat(array_with_padding(lna_output_TLA))
        ### End TLA ###
        
        return (xe_all, lna_all, Tm_all, lna_Tm_all)



    def post_Saha_expansion(self, starting_lna, args, threshold=1e-5):
        """
        Compute post-Saha expansion phase with two-photon corrections.

        Calculates ionization fraction including two-photon processes as
        perturbative corrections to Saha equilibrium until deviations
        exceed threshold.

        Parameters:
        -----------
        starting_lna : float
            Initial log scale factor
        args : tuple
            Background cosmology and cosmological parameters (BG, params)
        threshold : float, optional
            Threshold for deviation from Saha (default: 1e-5)

        Returns:
        --------
        tuple
            (xe_output, lna_output) - ionization fraction and log scale factor arrays
        """
        BG, params = args
        # Initial conditions
        TCMB = BG.TCMB(starting_lna, params)
        nH = BG.nH(starting_lna, params)
        xe0, _ = recomb_functions.xe_Saha(TCMB, nH)  # Saha equilibrium is our intial condition

        # Pre-allocate xe_output 
        xe_output = jnp.ones_like(self.concrete_axis_size)*jnp.inf
        lna_output = jnp.ones_like(self.concrete_axis_size)*jnp.inf
        iz = int(0)
        xe = xe0
        stop = False

        def compute_xe(carry):
            xe_output, lna_output, xe, iz, stop = carry

            lna = starting_lna + iz*self.integration_spacing

            # Cosmological parameters
            TCMB = BG.TCMB(lna, params)
            nH = BG.nH(lna, params)
            H = BG.H(lna, params)

            # Saha equilibrium for xe
            xe_Saha, s = recomb_functions.xe_Saha(TCMB, nH)
            dxe_Saha_dlna = -(cnst.rydberg / TCMB - 3./2.) * xe_Saha**2 / (2. * xe_Saha + s)

            # Compute xe using two-photon processes
            grad_dxedlna_func = grad(self.dxe_dlna_twophoton, argnums=0)
            grad_dxedlna = grad_dxedlna_func(xe_Saha, TCMB, TCMB, H, nH, 0.0)
            xe = xe_Saha + dxe_Saha_dlna / grad_dxedlna

            # Store current xe value in the output array
            xe_output = xe_output.at[iz].set(xe)
            lna_output = lna_output.at[iz].set(lna)

            # Check difference
            diff = jnp.abs(xe_Saha - xe)
            stop = diff > threshold  # Stop when diff < threshold

            # Increment index
            iz = iz + 1

            return (xe_output, lna_output, xe, iz, stop)

        def stop_condition(state):
            _, _, _, iz, stop = state
            return (iz < self.concrete_axis_size.size) & (~stop)  # Continue until stop condition is met or we run out of space

        # Initial state: (xe_output, xe, iz, stop flag)
        initial_state = (xe_output, lna_output, xe, iz, stop)

        # Run the while loop until the stop condition is met.
        # eqx.internal.while_loop with kind='checkpointed' installs a custom_vjp
        # so reverse-mode AD can traverse this dynamic-stop loop via treeverse
        # checkpointing. max_steps must be a static upper bound; the output
        # axis size serves that role here.
        final_state = eqx.internal.while_loop(
            stop_condition, compute_xe, initial_state,
            max_steps=self.concrete_axis_size.size,
            kind='checkpointed',
        )

        # Unpack the final state
        xe_output_final, lna_output_final, _, _, _ = final_state

        # Return the electron fraction array and the stopping `lna` value
        return xe_output_final, lna_output_final

    def xe_derivative_twophoton(self, lna, xe, args):
        """
        Compute ionization fraction derivative including two-photon processes.

        Derivative function for hydrogen ionization fraction evolution
        including two-photon transitions and correction functions.

        Parameters:
        -----------
        lna : float
            Log scale factor
        xe : float
            Current ionization fraction
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        float
            Time derivative dxe/dlna (units: dimensionless)
        """
        BG, params = args
    
        x1s = 1. - xe                # fraction of neutral hydrogen
        TCMB = BG.TCMB(lna, params)          # eV
        nH = BG.nH(lna, params)              # hydrogen number density, 1/cm^3
        H = BG.H(lna, params)                # Hubble parameter, 1/s
        GammaC = recomb_functions.Gamma_compton(xe, TCMB, params['YHe'])  # Compton scattering rate, 1/s

        Tm = TCMB * (1.-H/GammaC)

        Delta = self.get_current_correction_func(TCMB, args)
        dxedlna = self.dxe_dlna_twophoton(xe, TCMB, Tm, H, nH, Delta)

        return dxedlna

    def solve_emla_twophoton(self, lna_axis_init, lna_axis_final, xe0, args, rtol=1e-6, atol=1e-9,solver=Kvaerno3(),max_steps=4096):
        """
        Solve HYREC-2 EMLA evolution with two-photon processes.

        Integrates hydrogen recombination including effective multilevel atom
        approximation with two-photon transitions and correction functions.

        Parameters:
        -----------
        lna_axis_init : float
            Initial log scale factor
        lna_axis_final : float
            Final log scale factor
        xe0 : float
            Initial ionization fraction
        args : tuple
            Background cosmology and cosmological parameters (BG, params)
        rtol : float, optional
            Relative tolerance (default: 1e-6)
        atol : float, optional
            Absolute tolerance (default: 1e-9)
        solver : diffrax.Solver, optional
            ODE solver (default: Kvaerno3())
        max_steps : int, optional
            Maximum steps (default: 4096)

        Returns:
        --------
        tuple
            (xe_output, lna_output) - ionization fraction and log scale factor arrays
        """
        BG, params = args

        # Initial conditions
        TCMB_init = BG.TCMB(lna_axis_init, params)  # Initial CMB temperature
        initial_state = xe0
        term = ODETerm(self.xe_derivative_twophoton)

        t0 = lna_axis_init
        # t1 = lna_axis_final
        t1 = jnp.inf


        # don't want to double count the boundary lna, so start saving after one step
        t_arr = jnp.linspace(t0+self.integration_spacing, t0+2*max_steps*self.integration_spacing, 2*max_steps)

        save_at = SaveAt(ts=t_arr)
        adjoint=self.adjoint()

        def lna_check(t, y, args, **kwargs):
            return t > lna_axis_final

        event = Event(lna_check)

        # use diffrax default max_steps of 4096
        sol = diffeqsolve(
            term, solver, t0=t0, t1=t1, dt0=1e-3, 
            y0=initial_state, 
            args=args,
            stepsize_controller=PIDController(rtol, atol),saveat=save_at,
            event = event,
            adjoint=adjoint
        )
        
        xe_output = sol.ys
        lna_output = sol.ts

        return xe_output, lna_output

    def xe_tm_derivative(self, lna, state, args):
        """
        Compute coupled derivatives for ionization fraction and matter temperature.

        Derivative function for simultaneous evolution of hydrogen ionization
        fraction and matter temperature including Compton heating/cooling.

        Parameters:
        -----------
        lna : float
            Log scale factor
        state : array
            Current state [xe, Tm]
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        array
            Time derivatives [dxe/dlna, dTm/dlna] (units: dimensionless, eV)
        """
        xe, Tm = state
        BG, params = args
        
        TCMB = BG.TCMB(lna, params)          # eV
        nH = BG.nH(lna, params)              # hydrogen number density, 1/cm^3
        H = BG.H(lna, params)                # Hubble parameter, 1/s
        GammaC = recomb_functions.Gamma_compton(xe, TCMB, params['YHe'])  # Compton scattering rate, 1/s

        Delta = 0.0
        dxedlna = self.dxe_dlna_twophoton(xe, TCMB, Tm, H, nH, Delta)
        dTmdlna = (-2 * H * Tm + GammaC * (TCMB - Tm)) / H

        return jnp.array([dxedlna, dTmdlna])

    def solve_emla(self, lna0, xe0, args, rtol=1e-7,atol=1e-9,solver=Tsit5(),max_steps=4096):
        """
        Solve late-time EMLA evolution without two-photon processes.

        Integrates hydrogen recombination using effective multilevel atom
        approximation for late times when two-photon processes are negligible.

        Parameters:
        -----------
        lna0 : float
            Log scale factor at which initial xe is given
        xe0 : float
            Initial ionization fraction
        args : tuple
            Background cosmology and cosmological parameters (BG, params)
        rtol : float, optional
            Relative tolerance (default: 1e-7)
        atol : float, optional
            Absolute tolerance (default: 1e-9)
        solver : diffrax.Solver, optional
            ODE solver (default: Tsit5())
        max_steps : int, optional
            Maximum steps (default: 4096)

        Returns:
        --------
        tuple
            (xe_output, Tm_output, lna_output) - ionization fraction, matter temperature,
            and lna arrays
        """
        # Initial conditions
        t0 = lna0
        t1 = jnp.inf 

        BG,params = args

        # need to go at least twice max_steps to make sure we catch the t1 we actually want
        t_arr = jnp.linspace(t0+self.integration_spacing, t0+2*max_steps*self.integration_spacing, 2*max_steps)

        save_at = SaveAt(ts=t_arr) 
        
        TCMB_init = BG.TCMB(t0, params)  # Initial CMB temperature
        Tm0 = TCMB_init * (1.-BG.H(t0, params)/recomb_functions.Gamma_compton(xe0, TCMB_init, params['YHe']))

        initial_state = jnp.array([xe0, Tm0])
        term = ODETerm(self.xe_tm_derivative)
        adjoint=self.adjoint()

        def temperature_check(t, y, args, **kwargs):
            lna = t
            _, Tm = y
            TCMB = BG.TCMB(lna, params) 
            TR_MIN = recomb_functions.TR_MIN    # Minimum Tcmb in eV 
            T_RATIO_MIN = recomb_functions.T_RATIO_MIN  # Minimum Tratio 
            ratio = jnp.minimum(Tm / TCMB, TCMB / Tm)
            return jnp.logical_or(TCMB < TR_MIN, ratio < T_RATIO_MIN) # stop when true
        
        event = Event(temperature_check)

        # Call the solver
        sol = diffeqsolve(
            term, solver, t0=t0, t1=t1, dt0=1e-3, 
            y0=initial_state, 
            args=args,
            stepsize_controller=PIDController(rtol, atol),saveat=save_at,
            adjoint=adjoint,
            max_steps=max_steps,
            event = event
        )
        
        xe_output = sol.ys[:, 0] 
        Tm_output = sol.ys[:, 1] 

        return xe_output, Tm_output, sol.ts

    def dxe_dlna_twophoton(self, xe, TCMB, Tm, H, nH, Delta):
        """
        Compute two-photon recombination rate.

        Calculates ionization fraction evolution rate including two-photon
        transitions using effective multilevel atom approximation.

        Parameters:
        -----------
        xe : float
            Current ionization fraction
        TCMB : float
            CMB temperature (units: eV)
        Tm : float
            Matter temperature (units: eV)
        H : float
            Hubble parameter (units: s^{-1})
        nH : float
            Hydrogen number density (units: cm^{-3})
        Delta : float
            Correction function value

        Returns:
        --------
        float
            Recombination rate dxe/dlna (units: dimensionless)
        """

        x1s = 1.-xe

        # Interpolate transition rates
        A2s, A2p, B2s, B2p, _, _, R2p2s, R2s2p = recomb_functions.effective_coefficients(TCMB, Tm, H, nH, x1s)    

        # Compute the matrix and source vector forms of steady state equations, then solve the linear system for real and virtual populations.
        T, S = self.steady_state_equations(xe, H, nH, TCMB, A2s, A2p, B2s, B2p, R2p2s, R2s2p, Delta)
        X    = jnp.linalg.solve(T, S)
        x2s  = X[0]
        x2p  = X[1]

        return (x2s*B2s + x2p*B2p - xe**2*nH*(A2s+A2p)) / H

    def get_current_correction_func(self, TCMB, args):
        """
        Interpolate correction function for current cosmology.

        Interpolates correction function and applies cosmological parameter
        derivatives at current CMB temperature for accurate recombination rates.

        Parameters:
        -----------
        TCMB : float
            CMB temperature (units: eV)
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        float
            Correction function value (units: dimensionless)
        """
        # Fiducial cosmology values at which the correction functions were tabulated.
        omega_H_fid  = 0.01689
        omega_cb_fid = 0.14175
        Neff_fid     = 3.046

        BG, params = args


        # For the user inputed cosmology currently scanned over.
        omega_H  = params['omega_b']*(1-params['YHe'])
        omega_cb = params['omega_b'] + params['omega_cdm']

        Delta        = jnp.interp(TCMB, cnst.kB*self.swift[:, 0], self.swift[:, 1])
        dDelta_domcb = jnp.interp(TCMB, cnst.kB*self.swift[:, 0], self.swift[:, 2])
        dDelta_domH  = jnp.interp(TCMB, cnst.kB*self.swift[:, 0], self.swift[:, 3])
        dDelta_dNeff = jnp.interp(TCMB, cnst.kB*self.swift[:, 0], self.swift[:, 4])
        
        return Delta + (omega_cb-omega_cb_fid)*dDelta_domcb + (omega_H-omega_H_fid)*dDelta_domH + (params['Neff']-Neff_fid)*dDelta_dNeff

    def steady_state_equations(self, xe, H, nH, TCMB, A2s, A2p, B2s, B2p, R2p2s, R2s2p, Delta):
        """
        Set up steady-state level population equations.

        Constructs matrix equation for hydrogen level populations in
        steady-state approximation for 2s and 2p levels.

        Parameters:
        -----------
        xe : float
            Current ionization fraction
        H : float
            Hubble parameter (units: s^{-1})
        nH : float
            Hydrogen number density (units: cm^{-3})
        TCMB : float
            CMB temperature (units: eV)
        A2s : float
            2s recombination coefficient (units: cm^3 s^{-1})
        A2p : float
            2p recombination coefficient (units: cm^3 s^{-1})
        B2s : float
            2s photoionization coefficient (units: s^{-1})
        B2p : float
            2p photoionization coefficient (units: s^{-1})
        R2p2s : float
            2p→2s transition rate (units: s^{-1})
        R2s2p : float
            2s→2p transition rate (units: s^{-1})
        Delta : float
            Correction function value

        Returns:
        --------
        tuple
            (T, S) - transition matrix and source vector for level populations
        """
        T = jnp.zeros((2, 2), dtype="float64") 
        S = jnp.zeros(2, dtype="float64")

        x1s   = 1.-xe      # Recombined hydrogen fraction.
        
        # List of transition rates needed
        RLya  = self.Lyn_esc_rate(2, H, nH, x1s)  # Lyman-alpha escape rate.     
        R2s1s = 8.2206     # Two-photon transition rate from 2s to 1s.
        R2p1s = RLya / (1.+Delta) # Two-photon transition rate from 2p to 1s, with HYREC-2 fitting correction function.
        R1s2s = jnp.exp(-cnst.E21/TCMB)*R2s1s
        R1s2p = 3.*jnp.exp(-cnst.E21/TCMB)*R2p1s

        # Upper 2x2 part of T matrix.
        T = T.at[0, 0].set(B2s+R2s2p+R2s1s)
        T = T.at[0, 1].set(-R2p2s)
        T = T.at[1, 0].set(-R2s2p)
        T = T.at[1, 1].set(B2p+R2p2s+R2p1s)

        # First 2 entries of source vector elements
        S = S.at[0].set(xe**2*nH*A2s+x1s*R1s2s)
        S = S.at[1].set(xe**2*nH*A2p+x1s*R1s2p)

        return (T, S)

    def Lyn_esc_rate(self, n, H, nH, x1s):
        """
        Computes the Lyman-n escape rate, rate at which photons redshift past the Lyman-n line 
        without being absorbed. We use the convention that n=2 is Ly-alpha, n=3 is Ly-beta...

        Parameters
        ----------
        n : float
            Requested Lyman transition level, should be greater than 2.
        H : float
            Hubble parameter in s^-1.
        nH : float
            Hydrogen number density in cm^-3.
        x1s : float
            Fraction of 1s bound hydrogen.

        Returns
        -------
        RLyn : float
            Rate of escape of Lyman-n level.
        """
        lambda_lya = 2.*jnp.pi*cnst.hbar*cnst.c / cnst.rydberg * 4./3. # Lyman-alpha Wavelength
        RLya = 8.*jnp.pi*H/3./nH/x1s/lambda_lya**3      # Rate of escape of Lyman-alpha
        RLyn = (4*(n**2-1)/3/n**2)**3 * RLya            # (lambda_lya/lambda_lyn)^3 * RLya
        return RLyn

    def TLA_xe_deriv(self, lna, state, args):
        """
        Compute coupled derivatives for ionization fraction and matter temperature
        using Peebles three-level atom.

        Parameters:
        -----------
        lna : float
            Log scale factor
        state : array
            Current state [xe, Tm]
        args : tuple
            Background cosmology and cosmological parameters (BG, params)

        Returns:
        --------
        array
            Time derivatives [dxe/dlna, dTm/dlna] (units: dimensionless, eV)
        """
        xe, Tm  = state
        BG, params = args

        xHII = xe # since everything else is fully recombined
        nH = BG.nH(lna, params)
        TCMB = BG.TCMB(lna, params)          # eV
        H = BG.H(lna, params)  

        C = recomb_functions.peebles_C(jnp.exp(-lna) - 1.0, xHII, H, nH, args)
        alpha = recomb_functions.alpha_H(Tm)                     
        beta  = recomb_functions.beta_H(Tm)                  

        # dxe/d(lna) = (1/H) * dxe/dt
        dxe_dt = C * (beta * (1.0 - xe) - alpha * nH * xe**2)
        dxe_dloga = dxe_dt / H

        dTm_dloga = -2.0 * Tm + (recomb_functions.Gamma_compton(xe, TCMB, params['YHe']) / H) * (TCMB - Tm)

        return jnp.array([dxe_dloga, dTm_dloga])
    
    def solve_TLA(self, lna0, xe0, Tm0, args, rtol=1e-7, atol=1e-9, solver=Kvaerno3(), max_steps = 4096):
        """
        Solve late-time TLA evolution.

        Integrates hydrogen recombination using Peebles TLA in the region
        beyond where SWIFT corrections are tabulated.

        Parameters:
        -----------
        lna0 : float
            Starting log scale factor
        xe0 : float
            Initial ionization fraction
        Tm0: float
            Starting matter temperature
        args : tuple
            Background cosmology and cosmological parameters (BG, params)
        rtol : float, optional
            Relative tolerance (default: 1e-7)
        atol : float, optional
            Absolute tolerance (default: 1e-9)
        solver : diffrax.Solver, optional
            ODE solver (default: Tsit5())
        max_steps : int, optional
            Maximum steps (default: 4096)

        Returns:
        --------
        tuple
            (xe_output, Tm_output, lna_output) - ionization fraction, matter temperature,
            and log scale factor arrays
        """
        t0 = lna0
        t1 = jnp.inf # lna_axis.max
        # need to go at least twice max_steps to make sure we catch t1
        t_arr = jnp.linspace(t0+self.integration_spacing, t0+2*max_steps*self.integration_spacing, 2*max_steps)

        save_at = SaveAt(ts=t_arr) 
        # save_at = SaveAt(ts=lna_axis) # but start saving output at step 1 or later

        initial_state = jnp.array([xe0, Tm0])
        term = ODETerm(self.TLA_xe_deriv)
        adjoint=self.adjoint()

        def lna_check(t, y, args, **kwargs):
            return t > self.lna_end # stop when true
        
        event = Event(lna_check)

        sol = diffeqsolve(
            term, solver, t0=t0, t1=t1, dt0=1e-3, 
            y0=initial_state, 
            args=args,
            stepsize_controller=PIDController(rtol, atol),saveat=save_at,
            adjoint=adjoint,
            max_steps=max_steps,
            event=event
        )
        
        xe_output = sol.ys[:, 0]
        Tm_output = sol.ys[:, 1]

        return xe_output, Tm_output, sol.ts


import numpy as np
import jax.numpy as jnp
import equinox as eqx

from . import species

def load_specs(input_specs):

    specs = {}

    specs["use_LCDM_species"] = input_specs.get("use_LCDM_species", True)

    ### INPUT RELATED specs PARAMS ###
    # For reionization, input tau_reion the optical depth, or z_reion the hydrogen redshift?
    # WARNING: If the following parameter is set to True (by default) and the user inputs z_reion instead of tau_reion,
    # the default tau_reion will be used irrespective of the user input. The same is true if the user inputs tau_reion when
    # this is set to False. 
    specs["input_tau_reion"] = input_specs.get("input_tau_reion", True) 

    ### OUTPUT RELATED specs PARAMS ###
    specs["l_min"]     = input_specs.get("l_min", 2)
    specs["l_max"]     = input_specs.get("l_max", 2500)
    specs["lensing"]   = input_specs.get("lensing", False)
    specs["k_max"] = input_specs.get("k_max", 0.5)

    ### BBN ###
    specs["bbn_type"] = input_specs.get("bbn_type", "")
    specs["linx_reaction_net"] = input_specs.get("linx_reaction_net", "key_PRIMAT_2023")

    ### Boltzmann Hierarchy Cutoffs ###
    specs["l_max_g"]     = input_specs.get("l_max_g", 12)
    specs["l_max_pol_g"] = input_specs.get("l_max_pol_g", 10)
    specs["l_max_massless_nu"]    = input_specs.get("l_max_massless_nu", 17)
    specs["l_max_massive_nu"]  = input_specs.get("l_max_massive_nu", 17)

    ### Perturbation k-grid resolution ###
    specs["k_step_sub"]             = input_specs.get("k_step_sub", 5.e-2)
    specs["k_step_super"]           = input_specs.get("k_step_super", 2.e-3)
    specs["k_step_transition"]      = input_specs.get("k_step_transition", 2.e-1)
    specs["k_step_super_reduction"] = input_specs.get("k_step_super_reduction", 1.e-1)
    specs["k_min_tau0"]             = input_specs.get("k_min_tau0", 1.e-1)
    specs["k_max_tau0_over_l_max"]  = input_specs.get("k_max_tau0_over_l_max", 1.8)
    specs["H0_fid"]                 = input_specs.get("H0_fid", 2.255560e-04)
    specs["tau0_fid"]               = input_specs.get("tau0_fid",1.418668e+04)
    specs["rs_rec_fid"]             = input_specs.get("rs_rec_fid", 1.446279e+02)

    ### Transfer integration k-grid resolution ###
    specs["k_transfer_linstep"] = input_specs.get("k_transfer_linstep", 4.5e-1)
    specs["k_transfer_logstep"] = input_specs.get("k_transfer_logstep", 170.)
    specs["tau_rec_fid"]        = input_specs.get("tau_rec_fid", 281.040565)

    ### Pivot scale ###
    specs["k_pivot"]            = input_specs.get("k_pivot", 0.05)

    ### Set perturbations initial condition time ###
    specs["R_tc"] = input_specs.get("R_tc", 0.0015)
    specs["R_large"] = input_specs.get("R_large", 0.07)

    ### Perturbation Evolver Diffrax Settings ###
    specs["max_steps_PE"]    = input_specs.get("max_steps_PE", 2048)
    # Step size controller
    specs["k_split_PE"]      = input_specs.get("k_split_PE", 0.01)
    specs["rtol_small_k_PE"] = input_specs.get("rtol_small_k_PE", 1.e-5)
    specs["rtol_large_k_PE"] = input_specs.get("rtol_large_k_PE", 1.e-4)
    specs["atol_small_k_PE"] = input_specs.get("atol_small_k_PE", 1.e-10)
    specs["atol_large_k_PE"] = input_specs.get("atol_large_k_PE", 1.e-6)
    specs["pcoeff_PE"]       = input_specs.get("pcoeff_PE", 0.25)
    specs["icoeff_PE"]       = input_specs.get("icoeff_PE", 0.8)
    specs["dcoeff_PE"]       = input_specs.get("dcoeff_PE", 0.)

    ### Physical contributions to CMB temperature transfer function ###
    specs["scale_sw"]  = input_specs.get("scale_sw", 1)
    specs["scale_isw"] = input_specs.get("scale_isw", 1)
    specs["scale_dop"] = input_specs.get("scale_dop", 1)
    specs["scale_pol"] = input_specs.get("scale_pol", 1)

    # Preserve any unknown keys for custom species extensibility
    for key, value in input_specs.items():
        if key not in specs:
            specs[key] = value

    return specs

def populate_species(user_species, specs):
    species_list = ()
    species_dict = {}

    lcdm_species = (
        species.DarkEnergy,
        species.ColdDarkMatter,
        species.Baryon,
        species.Photon,
        species.MasslessNeutrino
    )

    i = 0
    diffrax_vector_idx = 1

    # Add baseline LCDM species if needed.
    if specs["use_LCDM_species"]:
        for s in lcdm_species:
            instance = s(diffrax_vector_idx, specs) # Creates an instance of s. init is now consistent across all species
            species_list = species_list + (instance,)
            species_dict[instance.name] = i

            i += 1
            diffrax_vector_idx += instance.num_moments

    if user_species is not None:
        for s in user_species:
            instance = s(diffrax_vector_idx, specs)
            species_list = species_list + (instance,)
            species_dict[instance.name] = i

            i += 1
            diffrax_vector_idx += instance.num_moments

    return species_list, species_dict

def get_k_axis_perturbations(specs):
    ks = np.zeros(2000)

    H0_fid     = specs["H0_fid"]
    tau0_fid   = specs["tau0_fid"]
    rs_rec_fid = specs["rs_rec_fid"]
    k_rec_fid  = 2.*jnp.pi/rs_rec_fid

    k_min = specs["k_min_tau0"] / tau0_fid
    k_max = specs["k_max_tau0_over_l_max"] / tau0_fid * specs["l_max"]

    k = k_min   
    ks[0] = k
    i = 0
    while k < k_max:
        step = (specs["k_step_super"]
                + 0.5 * (jnp.tanh((k-k_rec_fid)/k_rec_fid/specs["k_step_transition"])+1.)
                * (specs["k_step_sub"]-specs["k_step_super"])) * k_rec_fid

        scale2 = H0_fid**2

        step *= (k**2/scale2+1.)/(k**2/scale2+1./specs["k_step_super_reduction"])

        k += step
        i += 1
        ks[i] = k

    specs["k_min"]     = k_min
    specs["k_max_cmb"] = k

    # If lensing is needed, we need to extend max k by some amount to accurately compute high-l lensing.
    if specs["lensing"]:
        k_max = k + 0.3
        
        while k < k_max:
            step = 0.005

            k += step
            i += 1
            ks[i] = k

    # If the user specified a k_max above the current, we should add these as well.
    if k < specs["k_max"]:
        k_max = specs["k_max"]
        
        while k < k_max:
            step = 0.005

            k += step
            i += 1
            ks[i] = k

    ks = ks[np.where(ks>0)]
    k_axis_Pk_output = ks[np.where(ks<=specs["k_max"])]

    return jnp.array(ks), jnp.array(k_axis_Pk_output)

def get_k_axis_transfer(specs):
    ks = np.zeros(8000)

    k_period = 2*jnp.pi/(specs["tau0_fid"] - specs["tau_rec_fid"])

    k = specs["k_min"]
    ks[0] = k
    i = 0
    while k < specs["k_max_cmb"]:
        k = k \
            + k_period * specs["k_transfer_linstep"] * k \
            / (k + specs["k_transfer_linstep"]/specs["k_transfer_logstep"])
        i += 1
        ks[i] = k

    ks = jnp.array(ks[np.where(ks>0)])
    return ks


from jax import jit, config, lax, tree_util
import jax.numpy as jnp
from jaxtyping import Array
import numpy as np
import equinox as eqx

import diffrax
import jax

import sys
import os
file_dir = os.path.dirname(__file__)

from .hyrex import hyrex
from . import background, perturbations, spectrum, model_specs
from . import constants as cnst
from .ABCMBTools import bilinear_interp

from .linx.background import BackgroundModel
from .linx.abundances import AbundanceModel
from .linx.nuclear import NuclearRates
from .linx import const as linxconst
from .linx import thermo as linxThermo

config.update("jax_enable_x64", True)


class Model(eqx.Module):
    """
    Model configuration and computation manager.

    Creates instances of fluid species based on user input and organizes
    them for computation. Manages the full pipeline from background
    evolution through CMB power spectrum computation.

    Attributes:
    -----------
    PE : perturbations.PerturbationEvolver
        ABCMB perturbations module
    SS : spectrum.SpectrumSolver
        ABCMB spectrum module
    RecModel : hyrex.recomb_model
        HyRex recombination module
    specs : dict
        A dictionary of run options (expected to be static)
    species_list : tuple 
        A list of all fluids in the user cosmology
    species_dict : dict 
        A dictionary containing the names of all fluids, in the same order as 
        they appear in species_list.
    PArthENoPE_CLASS_table  : Array 
        A 2D table for interpolation of the helium-4 mass fraction based
        on the user's input baryon density and Neff
    thermo_model_DNeff : linx.BackgroundModel
        A LINX background model for BBN thermodynamics
    abundanceModel : linx.AbundanceModel
        A LINX abundance model used for computing the helium-4 mass fraction
        given the user's input baryon density, Neff, neutron lifetime, and
        nuclear reaction rates.

    Methods:
    --------
    __call__ : Compute CMB angular power spectra
    get_PTBG : Get perturbation table and background cosmology
    get_BG : Get background cosmology
    add_derived_parameters : Compute derived parameters
    """

    PE : perturbations.PerturbationEvolver
    SS : spectrum.SpectrumSolver
    RecModel : hyrex.recomb_model
    specs : dict

    species_list : tuple = ()
    species_dict : dict 
    
    PArthENoPE_CLASS_table  : Array 
    thermo_model_DNeff : BackgroundModel
    abundanceModel : AbundanceModel

    adjoint : "diffrax.adjoint" = eqx.field(static=True)

    ### ADDING SPECIES: add has_ parameter and add condition to append to tuple.
    # In the init, all species that are present within the model should be set to True.
    # All couplings present between species should be set to true. 
    def __init__(self,
                 user_species=None,
                 **kwargs
                 ):
        """
        Initialize Model instance.

        Sets up fluid species, recombination model, and spectrum solver
        based on configuration parameters.

        Parameters:
        -----------
        user_species : tuple
            A tuple of user-defined fluids to be included in the cosmology
        **kwargs : dict
            Configuration options passed as keyword arguments.
            Any unknown keys will be preserved for custom species extensibility.
        """

        # Pull adjoint out of kwargs before load_specs — it must NOT end up
        # inside self.specs (a non-JAX pytree leaf breaks lax.cond / filter_jit
        # tracing). Default preserves prior ForwardMode behavior.
        adjoint = kwargs.pop("adjoint", diffrax.ForwardMode)

        # Fill in all user defined and missing specs parameters
        specs = model_specs.load_specs(kwargs)
        self.specs = specs

        # Populate all species
        self.species_list, self.species_dict = model_specs.populate_species(
            user_species,
            specs,
        )   

        # Initialize perturbation evolver
        k_axis_perturbations, k_axis_Pk_output = model_specs.get_k_axis_perturbations(specs)
        self.PE = perturbations.PerturbationEvolver(
            self.species_list,
            self.species_dict,
            k_axis_perturbations,
            specs,
            adjoint=adjoint,
        )

        # Intialize spectrum solver
        k_axis_transfer = model_specs.get_k_axis_transfer(specs)
        self.SS = spectrum.SpectrumSolver(
            specs["l_min"],
            specs["l_max"],
            specs["lensing"],
            k_axis_transfer,
            k_axis_Pk_output,
            k_pivot=specs["k_pivot"],
            scale_sw=specs["scale_sw"],
            scale_isw=specs["scale_isw"],
            scale_dop=specs["scale_dop"],
            scale_pol=specs["scale_pol"]
        )

        # Initialize recombination model
        self.RecModel = hyrex.recomb_model(adjoint=adjoint) # DO NOT CHANGE z1 FROM 0

        # Initialize BBN model
        self.PArthENoPE_CLASS_table = jnp.asarray(np.loadtxt(file_dir+'/sBBN_2025_CLASS.txt'))
        # initialize LINX
        if self.specs["bbn_type"].lower() == "linx":
            self.thermo_model_DNeff = BackgroundModel(adjoint=adjoint)
            self.abundanceModel = AbundanceModel(NuclearRates(nuclear_net=self.specs["linx_reaction_net"]), adjoint=adjoint)
        else:
            self.thermo_model_DNeff = None
            self.abundanceModel = None

        self.adjoint = adjoint

    # need this outside of the jit context
    # since we want LINX to run on CPU
    def __call__(self, params : dict = {}):
        """
        Compute CMB angular power spectra for given parameters.

        Runs the full pipeline from background evolution through
        perturbation integration to CMB power spectrum computation.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        tuple
            (ℓ values, (C_ℓ^TT, C_ℓ^TE, C_ℓ^EE)) for computed multipoles
        """

        
        full_params = self.add_derived_parameters(params)
        return self.run_cosmology_abbr(full_params)
        
    ### JITTED OR JITTABLE FUNCTIONS ###

    @eqx.filter_jit
    def run_cosmology_abbr(self, params : dict):
        """
        Compute CMB angular power spectra for given parameters.

        Runs the full pipeline from background evolution through
        perturbation integration to CMB power spectrum computation.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        tuple
            (ℓ values, (C_ℓ^TT, C_ℓ^TE, C_ℓ^EE)) for computed multipoles
        """

        # let the user know the code is compiling
        print("")
        print('              /\\  ')
        print('             /  \\   ')
        print('            / /\\ \\  ')
        print('           / /__\\ \\    ___   ___  ')
        print('          / ______ \\  | _ \\ / __\\ _  _  ')
        print('         / /      \\ \\ |  _// /   | \\/ | __  ')
        print('        / /        \\ \\| _ \\\\ \\___||\\/||| -)  ')
        print('       /_/          \\_|___/ \\___/||  |||_-) is compiling...')
        print('\\_____/      ')
        print("")

        # Compute background and linear perturbations
        PT, BG = self.get_PTBG(params)

        # Compute CMB power spectra
        Cls = self.SS.get_Cl(PT, BG, params)
        l = self.SS.ells
        
        # Compute linear matter power spectrum
        Pk = self.SS.Pk_lin(self.SS.k_axis_Pk_output, 0., PT, params)
        k = self.SS.k_axis_Pk_output

        # Package
        output = Output(
            Cls[0], Cls[1], Cls[2], Pk,
            l, k, BG, PT, params
        )

        return output

    @eqx.filter_jit
    def get_PTBG(self, params : dict):
        """
        Get perturbation table and background.

        Computes background and evolves perturbations for the given parameters.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        tuple
            (PerturbationTable, Background) objects
        """
        BG = self.get_BG(params)
        PT = self.PE.full_evolution((BG, params))

        return PT, BG

    @eqx.filter_jit
    def get_BG(self, params : dict):
        """
        Get background for given parameters.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        background.Background
            Background object
        """
        # Bind to a local so both closures capture a plain class rather than
        # an attribute lookup on self. The class is never placed in the
        # lax.cond operand tuple (keeping it a valid JAX pytree).
        adjoint = self.adjoint
        def get_BG_z_reion(args):
            params, species_list, RecModel = args
            return background.Background(params, species_list, RecModel, background.ReionizationModelFromZ, adjoint=adjoint)

        def get_BG_tau_reion(args):
            params, species_list, RecModel = args
            return background.Background(params, species_list, RecModel, background.ReionizationModelFromTau, adjoint=adjoint)

        BG = lax.cond(
            self.specs["input_tau_reion"],
            get_BG_tau_reion,
            get_BG_z_reion,
            (params, self.species_list, self.RecModel)
        )
        
        return BG

    def add_derived_parameters(self, param_in : dict) -> dict:
        # we do not want to do in-place updates so we can
        # recycle dicts if LINX option is used
        params = param_in.copy()

        # Default parameters except Neff and YHe
        params['h']             = jnp.array(params.get('h', 0.6736))
        params['H0']            = jnp.array(params['h'] * cnst.H0_over_h)
        params['omega_cdm']     = jnp.array(params.get('omega_cdm', 0.120))
        params['omega_b']       = jnp.array(params.get("omega_b", 0.02237))
        params['A_s']           = jnp.array(params.get('A_s', 2.1e-9))
        params['n_s']           = jnp.array(params.get('n_s', 0.9649))
        params['TCMB0']         = jnp.array(params.get('TCMB0', 2.34865418e-4))

        # Reionization
        if self.specs["input_tau_reion"]:
            params['tau_reion'] = jnp.array(params.get('tau_reion', 0.0544))
        else:
            params['z_reion'] = jnp.array(params.get('z_reion', 7.67))
        params['Delta_z_reion'] = jnp.array(params.get('Delta_z_reion', 0.5))
        params['z_reion_He']    = jnp.array(params.get('z_reion_He', 3.5))
        params['Delta_z_reion_He'] = jnp.array(params.get('Delta_z_reion_He', 0.5))
        params['exp_reion']     = jnp.array(params.get('exp_reion',1.5))

        # Here we fill in a fake omega_Lambda just so that the DE energy density can be computed in a loop.
        # This fake quantity will not be used in anything, and later the correct omega_Lambda will be computed.
        # Purely computational, no physics used or messed up.
        params['omega_Lambda'] = 0.

        # Massive neutrinos
        params['T_nu_massive']  = jnp.array(params.get('T_nu_massive', 0.71611)) # Massive neutrino temperature, as a ratio to TCMB.
        params['N_nu_massive']  = jnp.array(params.get('N_nu_massive', 0))  # Number of massive neutrinos
        params['m_nu_massive']  = jnp.array(params.get('m_nu_massive', 0.06)) # Massive neutrino mass, in eV

        ### CHECKING INPUT COMPATIBILITY ###

        input_N    = params.get('N_nu_massless') != None
        input_Neff = params.get('Neff') != None
        input_T_nu_massless = params.get('T_nu_massless') != None

        # If the user input both massless neutrino number and Neff, throw an error. Our code treats these as 1-to-1, see paper.
        if input_N and input_Neff:
            print("You can only input one of N_nu_massless or Neff, but got values N_nu_massless={} and Neff={}.".format(params["N_nu_massless"], params["Neff"]))
            sys.exit()

        # If the user input either N_massless or Neff, but requested LINX, throw an error. LINX will compute the correct values.
        if (input_N or input_Neff or input_T_nu_massless) and self.specs["bbn_type"].lower() == "linx":
            print(
                "You have specified a value for N_nu_massless and/or Neff and/or T_nu_massless, \n"
                "but LINX instead expects a parameter 'Delta_Neff_init' which will be used to \n" \
                "compute Neff. Refer to LINX docs or https://arxiv.org/abs/2408.14538 for more info.\n" \
            )
            sys.exit()

        if not input_N and not input_Neff and self.specs["bbn_type"].lower() != "linx":
            params["N_nu_massless"] = 3 - params['N_nu_massive']
            input_N = True

        ### END OF INPUT COMPATIBILITY ###
        # now that we have verified the user put in the right parameters we can set T_nu_massless
        params['T_nu_massless'] = jnp.array(params.get('T_nu_massless', 0.71636856)) # Massless neutrino temperature, as a ratio to TCMB

        ### HELIUM FRACTION AND Neff ###
        # Regardless of bbn_type, these two parameters will be set by the end.

        lna_early = -23.
        a_early = jnp.exp(lna_early)

        # Case 1: The user specifies the true number of massless neutrinos. Note this is distinct from CLASS' N_ur which
        # is computed assuming T_massless = (4/11)^(1/3) x T_CMB.
        # Here, Neff will be inferred from the cosmological fluid content. 
        # In particular if the universe contains massive neutrinos, we account for the error incurred when using a late time
        # massive neutrino temperature which underestimates the massive neutrino energy density at early time, when Neff is set.
        # We account for this by adding the missing relativistic energy in massive neutrinos at early times to the massless fluid.
        # See detail in paper.
        if input_N:
            rho_g = 0.
            rho_nu = 0.
            rho_extra = 0.
            for s in self.species_list:
                rho = s.rho(lna_early, params)
                if s.name == "Photon":
                    rho_g += rho
                elif "neutrino" in s.name.lower():
                    rho_nu += rho
                else:
                    rho_extra += rho

            Neff_raw     = (rho_nu+rho_extra)/rho_g * (8./7.) * (11./4.)**(4./3.) # Uncorrected Neff using T_nu_massive today
            rho_nu_early = 7/8 * (params["N_nu_massless"] + params["N_nu_massive"]) * params["T_nu_massless"]**4 * rho_g # Correct using massless neutrino temp.
            params["Neff"] = (rho_nu_early+rho_extra)/rho_g * (8./7.) * (11./4.)**(4./3.)
            params["N_nu_massless"] = params["N_nu_massless"] + params["Neff"] - Neff_raw # Add difference to massless sector.

        if self.specs["bbn_type"].lower() == "table":
            # Applies if user requested BBN table to be user.
            # In this case Neff must already have been set, and can be used to interp YHe.

            # interpolate CLASS ParthENoPE table
            bbn = self.PArthENoPE_CLASS_table
            omegab_all = bbn[:, 0]
            DNeff_all = bbn[:, 1]
            YHe_all = bbn[:, 2]

            # we have to hardcode these values to be jit safe (alternatively we 
            # could read them in at runtime, but these tables don't update 
            # frequently)
            n2 = 13 
            n1 = 701

            omegab = omegab_all[:n1]
            DNeff = DNeff_all[::n1] 

            YHe_grid = YHe_all.reshape(n2, n1)
            
            # Neff = params["Neff"] # less extensible option
            a_bbn = cnst.TCMB_today*1e-6/0.01   # neutrino decoupling is well over by 10 keV, so 
                                                # compute Neff at a scale factor approximately 
                                                # corresponding to this temperature
            lna_bbn = jnp.log(a_bbn)

            # Comprehensive Neff, includes all relativsitic species at early times.
            Neff_BBN = params["Neff"]
            
            # last two args are user input omega_b and (Neff_BBN - 3.046) (MUST be 3.046 as 
            # this was assumed when constructing the PArthENoPE table)
            res_YHe = bilinear_interp(omegab, DNeff,YHe_grid, params['omega_b'],Neff_BBN - 3.046)

            # tabulated result is Yp_CMB
            params['YHe'] = res_YHe
        elif self.specs["bbn_type"].lower() == "linx":
            # Applies if user requested to run LINX.
            # For this branch to happen, Neff must NOT have already been set. 
            # Logic above has already accounted for this, since input_T and input_Neff must both be False
            # for LINX to execute.
            params['Delta_Neff_init'] = jnp.array(params.get('Delta_Neff_init', 0.))
            (
                t_vec_ref, a_vec_ref, rho_g_vec, rho_nu_vec, rho_NP_vec, P_NP_vec, Neff_vec 
            ) = eqx.filter_jit(self.thermo_model_DNeff,backend='cpu')(params['Delta_Neff_init'])

            # convert user input omega_b to eta_fac LINX expects
            eta_fac = params['omega_b'] * linxconst.Omegabh2_to_eta0/linxconst.eta0

            abundances = eqx.filter_jit(self.abundanceModel,backend='cpu')(
                rho_g_vec,
                rho_nu_vec,
                rho_NP_vec,
                P_NP_vec,
                t_vec=t_vec_ref,
                a_vec=a_vec_ref,  
                eta_fac = eta_fac,
                tau_n_fac = jnp.asarray(params.get("tau_n_fac", 1.0)),
                nuclear_rates_q = jnp.asarray( params.get("nuclear_rates_q", jnp.zeros( len(self.abundanceModel.nuclear_net.reactions) )) )
                )
  
            # number abundance
            try:
                params['T_nu_massless'] = jax.device_put(
                    linxThermo.T_nu(rho_nu_vec[-1]) / linxThermo.T_g(rho_g_vec[-1]),
                    device=jax.devices('gpu')[0]
                )
                params['Neff'] = jax.device_put(Neff_vec[-1],device=jax.devices('gpu')[0])
                YHe_BBN = jax.device_put(4*abundances[5],device=jax.devices('gpu')[0])
            except: # no GPU
                params['T_nu_massless'] = linxThermo.T_nu(rho_nu_vec[-1]) / linxThermo.T_g(rho_g_vec[-1])
                params['Neff'] = Neff_vec[-1]
                YHe_BBN = 4*abundances[5]
                pass
        
            # CMB uses real mass fraction
            Yp_CMB = 1./(4*cnst.mH/cnst.mHe*(1/YHe_BBN - 1) + 1)
            params['YHe'] = Yp_CMB

            # Now Neff has been set by LINX but massless neutrino number has yet to be calculated.
            # we now set the input_Neff flag to True so the branch below takes care of this.
            input_Neff = True
        else:
            # Applies if user wanted neither LINX or BBN table. 
            params['YHe'] = jnp.array(params.get('YHe', 0.245))


        # Case 2: User specifies the total Neff of the universe, including neutrinos and all other relativistic species at early times.
        # Then we subtract off all relativistic energy densities from Neff and assign the remaining to massless neutrinos.
        # Since massless neutrino temperature is already specified here, the true derived parameter is N_nu_massless, the physical
        # number of massless neutrinos.
        # The philosophy is that if we're increasing Neff, we are not heating the existing neutrinos, we are adding extra neutrinos
        # at the same temperature. At the CMB level these are indistinguishable, but we chose the later convention. 
        # Note, if after the deduction there's not enough energy density for massless neutrinos (N_nu_massless < 0), ABCMB throws an error.
        if input_Neff:
            rho_g = 0.
            rho_extra = 0.
            for s in self.species_list:
                if s.name == "Photon":
                    rho = s.rho(lna_early, params)
                    rho_g += rho
                elif s.name != "MasslessNeutrino":
                    rho = s.rho(lna_early, params)
                    rho_extra += rho
            rho1nu = 7/8 * (4/11)**(4/3) * rho_g
            
            params['N_nu_massless'] = (params["Neff"] - rho_extra/rho1nu) * ((4/11)**(1/3) / params["T_nu_massless"])**4

        # Loop over matter fluids to compute total matter density today.
        rho_m = 0.
        for s in self.species_list:
            if s.is_matter:
                rho_m += s.rho(0., params)
        params['omega_m']      = rho_m / (3 * cnst.H0_over_h**2/8/jnp.pi/cnst.G) # Fractional matter density
        params['R_b']          = params['omega_b'] / params['omega_m'] # Baryon fraction
    
        # Loop over all fluids and compute energy density at very early time, inferring radiation energy density this way.
        a_early = jnp.exp(-23.)
        rho_r  = 0.
        rho_nu = 0.
        for s in self.species_list:
            rho_r += s.rho(jnp.log(a_early), params)
            if "neutrino" in s.name.lower():
                rho_nu += s.rho(jnp.log(a_early), params)

        params['omega_r']      = rho_r * a_early**4 / (3 * cnst.H0_over_h**2/8/jnp.pi/cnst.G) # Fractional radiation density today
        params['R_nu']         = rho_nu / rho_r # Fractional radiation density in neutrinos, defined at early times. Used for setting adiabatic ICs.

        # Special density parameter defined for computing adiabatic initial conditions
        # Defined as Omega_m / sqrt{Omega_r} * H0, in units of 1/Mpc
        params['om'] = params['omega_m'] / jnp.sqrt(params['omega_r']) * cnst.H0_over_h / cnst.c_Mpc_over_s

        # Having inferred correct omega_m and omega_r, compute correct omega_Lambda
        params['omega_Lambda'] = params['h']**2 - params['omega_r'] - params['omega_m']

        # There is NO NEED to modify this list!!  This is to make sure any new
        # user-defined keys will not trigger recompilation by wrapping them in
        # jnp.array, as is done manually above for all other keys.  LINX-
        # related inputs are intentionally excluded from this list!
        expected_keys = {
            'h', 'H0', 'omega_cdm', 'omega_b', 'A_s', 'n_s', 'TCMB0',
            'tau_reion', 'z_reion', 'Delta_z_reion', 'z_reion_He', 'Delta_z_reion_He', 'exp_reion',
            'omega_Lambda', 'T_nu_massive', 'N_nu_massive', 'm_nu_massive',
            'N_nu_massless', 'Neff', 'T_nu_massless', 'YHe',
            'omega_m', 'R_b', 'omega_r', 'R_nu', 'om'
        }
        
        for key, value in param_in.items():
            if key not in expected_keys:
                params[key] = jnp.array(value)

        return params

class Output(eqx.Module):
    """
    Object containing final and intermediate results from one cosmological simulation.

    Attributes:
    -----------
    ClTT : jnp.array
        Temperature-temperature power spectrum
    ClTE : jnp.array
        Temperature-polarization power spectrum
    ClEE : jnp.array
        Polarization-polarization power spectrum
    Pk : jnp.array
        Matter power spectrum
    l : jnp.array
        Multipoles l at which ClTT/ClTE/ClEE are output
    k : jnp.array
        Wavenumbers k at with Pk is output
    BG  : background.Background
        Background object containing functions like Hubble, recombination history, etc
    PT : perturbations.PerturbationTable
        Perturbation table including perturbations for all fluids
    params : dict
        Complete parameter dictionary including derived parameters
    """

    # Power spectra
    ClTT : jnp.array
    ClTE : jnp.array
    ClEE : jnp.array
    Pk   : jnp.array

    l  : jnp.array
    k  : jnp.array
    BG : background.Background
    PT : perturbations.PerturbationTable
    params : dict



import jax
from jax import config, vmap, lax
import numpy as np
import jax.numpy as jnp
import equinox as eqx
from diffrax import diffeqsolve, ODETerm, Kvaerno5, Tsit5, SaveAt, PIDController, ForwardMode
import optimistix as optx

from .hyrex.array_with_padding import array_with_padding
from .hyrex import recomb_functions
from . import ABCMBTools as tools
from . import constants as cnst

import os
file_dir = os.path.dirname(__file__)
config.update("jax_enable_x64", True)

class Background(eqx.Module):
    """
    Background cosmology module for cosmological calculations.

    Computes background quantities including Hubble parameter, conformal time,
    recombination history, and optical depth evolution.

    Attributes:
    -----------
    species_list : tuple
        A list of all fluids in the cosmology
    lna_tau_tab : jnp.array
        Log scale factor axis used to tabulate conformal time
    tau_tab : jnp.array 
        Tabulated conformal time. 
    tau0 : float 
        Conformal time today in Mpc.
    xe_tab : array_with_padding
        Tabulated free electron fraction xe during recombination
    lna_xe_tab : array_with_padding
        Log scale factor axis corresponding to tabulated xe values.
    Tm_tab : array_with_padding
        Tabulated matter temperature Tm during recombination
    lna_Tm_tab : array_with_padding
        Log scale factor axis corresponding to tabulated Tm values.
    kappa_func : diffrax.solution
        Visibility function
    z_reion : float
        Redshift of hydrogen reionization in the CAMB parameterization.
    tau_reion : float 
        Optical depth to reionization
    lna_rec : float
        Log scale factor of recombination
    rA_rec : float 
        Comoving angular diameter distance at recombination in Mpc
    rs_d : float 
        Sound horizon at baryon decoupling in Mpc
    z_d : float
        Redshift of baryon decoupling
    lna_transfer_start : float 
        Log scale factor at which to begin integrating transfer functions.
    lna_visibility_stop : float 
        Log scale factor at which to stop integrating T1, T2, and E sources due to small visibility functions. 
        Only used for l<400.

    Recombination Unrelated Methods:
    --------------------------------
    rho_tot : Compute total energy density (units: eV cm^{-3})
    P_tot : Compute total pressure (units: eV cm^{-3})
    H : Compute Hubble parameter (units: s^{-1})
    aH : Compute conformal Hubble parameter (units: Mpc^{-1})
    aH_prime : Compute derivative of conformal Hubble (units: Mpc^{-1})
    d2adtau2_over_a : Compute second derivative of scale factor (units: Mpc^{-2})
    tau : Compute conformal time (units: Mpc)
    z_d : Compute baryon decoupling redshift (units: dimensionless)
    rs_d : Compute sound horizon at decoupling (units: Mpc)

    Recombination Related Methods:
    ------------------------------
    xe : Compute free electron fraction (units: dimensionless)
    Tm : Compute matter temperature (units: eV)
    mu_bar : Compute mean molecular mass (units: eV)
    cs2 : Compute baryon sound speed squared (units: dimensionless)
    nH : Compute hydrogen number density (units: cm^{-3})
    TCMB : Compute CMB temperature (units: eV)
    tau_c : Compute Thomson scattering time (units: Mpc)
    kappa : Compute optical depth (units: dimensionless)
    visibility : Compute visibility function (units: Mpc^{-1})
    """

    # params : dict
    species_list : tuple
 
    lna_tau_tab = jnp.linspace(-33.0, 0.0, 10000) # Axis for tabulating conformal time.
    tau_tab : jnp.array                     # Tabulated conformal time. 
    tau0 : float # Conformal time today

    # Recombination related
    xe_tab     : "array_with_padding"
    lna_xe_tab : "array_with_padding"
    Tm_tab     : "array_with_padding"
    lna_Tm_tab : "array_with_padding"
    kappa_func : "diffrax.solution"
    z_reion    : float
    tau_reion  : float 
    lna_rec    : float
    rA_rec     : float # Comoving angular diameter distance at recombination.

    # Transfer related
    lna_transfer_start : float # Time where transfer functions start integrating.
    lna_visibility_stop : float # Time to stop integrating T1, T2, and E sources due to small visibility functions. Only used for l<400

    adjoint : "diffrax.adjoint" = eqx.field(static=True)

    def __init__(self, params, species_list, RecModel, ReionModel, adjoint=ForwardMode):
        """
        Initialize Background cosmology module.

        Computes and tabulates conformal time, recombination history,
        optical depth, and key cosmological epochs.

        Parameters:
        -----------
        params : dict
            Cosmological parameters
        species_list : tuple
            List of fluid species for energy density calculations
        RecModel : callable
            Recombination module for computing xe and Tm histories
        ReionModel : callable
            Reionization module for computing the xe correction.
        """
        self.adjoint = adjoint
        # self.params = params
        self.species_list = species_list

        self.tau_tab = self._tabulate_conformal_time(params)
        self.tau0 =self.tau(0.)

        self.run_recombination(params, RecModel, ReionModel)

        # Find approximate maximum of visibility function.
        lna_vals = jnp.linspace(-8.0, -4.0, 1500) # Decoupling should have happened at some time in this interval.
        vis_vals = vmap(self.visibility,in_axes=[0,None])(lna_vals, params)
        self.lna_rec = lna_vals[jnp.argmax(vis_vals)]
        self.lna_visibility_stop = lna_vals[jnp.argmin((vis_vals - 1.e-3)**2)]
        self.rA_rec = self.tau0 - self.tau(self.lna_rec)

        # Find approximate early time when aH x tau_c = 0.008
        lna_vals = jnp.linspace(-15.0, -6.0, 5000)
        aH_tau_c_vals = vmap(self.aH,in_axes=[0,None])(lna_vals,params)*self.tau_c(lna_vals,params)
        self.lna_transfer_start = lna_vals[jnp.argmin((aH_tau_c_vals-0.008)**2)]

    def run_recombination(self, params, RecModel, ReionModel):
        """
        Call HyRex to get the primary recombination history, then patch on tanh reionization.
        For now assumes the user will specify tau_reio. 
        """
        ### RECOMBINATION ###

        # Run hyrex to tabulate recombination output
        xe, self.lna_xe_tab, self.Tm_tab, self.lna_Tm_tab = RecModel((self,params))

        ### REIONIZATION ###
        reion_model = ReionModel(self, params)
        self.z_reion = reion_model.z_reion
        self.tau_reion = reion_model.tau_reion

        xe_reion_correction = reion_model.xe_reion(self.lna_xe_tab.arr, self.z_reion, params)
        xe_full_arr = xe_reion_correction + xe.arr 
        xe_full = array_with_padding(xe_full_arr)

        self.xe_tab = xe_full

        self.kappa_func = self._tabulate_optical_depth(params)

    def rho_tot(self, lna, params):
        """
        Compute total energy density.

        Sums energy density over all species in the universe.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Total energy density (units: eV cm^{-3})

        Notes:
        ------
        User should not modify this function without careful consideration.
        """
        rho_tot = 0.
        for i in range(len(self.species_list)):
            rho_tot += self.species_list[i].rho(lna, params)
        return rho_tot
    
    def P_tot(self, lna, params):
        """
        Compute total pressure.

        Sums pressure over all species in the universe.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Total pressure (units: eV cm^{-3})

        Notes:
        ------
        User should not modify this function without careful consideration.
        """
        P_tot = 0.
        for i in range(len(self.species_list)):
            P_tot += self.species_list[i].P(lna, params)
        return P_tot

    def H(self, lna, params):
        """
        Compute Hubble parameter.

        Uses Einstein equation H = sqrt(8πG/3 ρ_tot) to account for
        novel species without well-defined density parameters.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Hubble parameter (units: s^{-1})
        """
        return jnp.sqrt(8.*jnp.pi*cnst.G*self.rho_tot(lna, params)/3.)

    def aH(self, lna, params):
        """
        Compute conformal Hubble parameter.

        Calculates conformal Hubble H = a*H = da/dτ where τ is conformal time.
        Uses Mpc units for perturbation calculations.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Conformal Hubble parameter (units: Mpc^{-1})
        """
        return jnp.exp(lna)*self.H(lna, params) / cnst.c_Mpc_over_s
    
    def aH_prime(self, lna, params):
        """
        Compute derivative of conformal Hubble parameter.

        Uses second Friedmann equation to compute d(aH)/d(ln a).
        See Eq.(20) of arXiv:9506072.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Derivative of conformal Hubble (units: Mpc^{-1})
        """
        return -4.*jnp.pi*cnst.G*jnp.exp(lna)**2/3./self.aH(lna, params) * (self.rho_tot(lna,params)+3.*self.P_tot(lna, params)) / cnst.c_Mpc_over_s**2

    def d2adtau2_over_a(self, lna, params):
        """
        Compute second derivative of scale factor.

        Calculates d²a/dτ²/a where τ is conformal time.
        Appears in perturbation evolution equations.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Second derivative of scale factor (units: Mpc^{-2})
        """

        return self.aH(lna, params)**2 + self.aH(lna, params)*self.aH_prime(lna, params)
    
    def _dtau_dlna(self, lna, y, args):
        """
        Compute derivative of conformal time with respect to ln(a).

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        y : float
            Current conformal time value
        args : dict
            Cosmological parameters

        Returns:
        --------
        float
            Derivative dτ/d(ln a) (units: Mpc)
        """
        params = args
        return 1./self.aH(lna, params)

    def _tabulate_conformal_time(self, params):
        """
        Tabulate conformal time as function of ln(a).

        Integrates dτ/d(ln a) = 1/aH from early times to today
        using radiation-dominated initial conditions. We stitch an
        analytic early-time solution to a Diffrax dense
        interpolation, taking care not to evaluate out of bounds.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        array
            Tabulated conformal time values (units: Mpc)
        """

        lna_cut = -16.1  # use analytic approx before this
        # Analytic early-time approximation
        tau_approx = lambda lna: (
            jnp.exp(lna) / (cnst.H0_over_h / cnst.c_Mpc_over_s) / jnp.sqrt(params["omega_r"])
        )

        lna_end = self.lna_tau_tab[-1]

        # ---- Diffrax solve (dense interpolation) ----
        term = ODETerm(self._dtau_dlna)
        controller = PIDController(rtol=1e-8, atol=1e-8)
        saveat = SaveAt(dense=True)
        adjoint=self.adjoint()

        sol = diffeqsolve(
            term,
            solver=Kvaerno5(),
            t0=lna_cut,
            t1=lna_end,
            dt0=1e-5,
            y0=tau_approx(lna_cut),
            saveat=saveat,
            stepsize_controller=controller,
            args=params,
            adjoint=adjoint,
        )

        # Numerical jitter causes this interpolation to go out of bounds on
        # some machines, so we do some extra work to safeguard that here:

        # Strictly inside [lna_cut, lna_end); avoid touching internal sol.ts (may be None).
        # nextafter gets the next representable float below lna_end to ensure in-bounds.
        lna_hi = jnp.nextafter(lna_end, -jnp.inf)

        def _tau_from_sol(l):
            l_in = jnp.clip(l, lna_cut, lna_hi)
            return sol.evaluate(l_in)

        def _tau_combined(l):
            # cond is faster than where since untaken branch is not evaluated
            return lax.cond(l > lna_cut, _tau_from_sol, tau_approx, l)

        tau_tab = vmap(_tau_combined)(self.lna_tau_tab)

        # Replace any remaining non-finite entries with analytic fallback
        tau_tab = jnp.where(jnp.isfinite(tau_tab), tau_tab, vmap(tau_approx)(self.lna_tau_tab))

        return tau_tab

    def tau(self, lna):
        """
        Compute conformal time.

        Interpolates from pre-tabulated conformal time history.
        Conformal time τ satisfies dτ = dt/a where t is cosmic time.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor

        Returns:
        --------
        float
            Conformal time (units: Mpc)

        Notes:
        ------
        IDEA: Make Background a repeatedly initiated module with both
        species_list and params stored. Upon initiation, a full history
        of conformal time is calculated with diffrax and stored for
        interpolation. This can be done by approximating early time with
        radiation approximation, and starting diffrax integration at the
        early time with appropriate initial conditions.
        """

        return tools.fast_interp(lna, self.lna_tau_tab[0], self.lna_tau_tab[-1], self.tau_tab)

    ### RECOMBINATION RELATED ###

    def xe(self, lna):
        """
        Compute free electron fraction.

        Interpolates from pre-tabulated recombination history with
        boundary conditions for early and late times.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor

        Returns:
        --------
        float
            Free electron fraction (units: dimensionless)

        Notes:
        ------
        The logic flow is equivalent to:

        if lna < self.lna_xe_tab.arr[0]: return self.xe_tab[0]

        elif lna > self.lna_xe_tab.lastval: return self.xe_tab.lastval

        else: return jnp.interp(lna, self.lna_xe_tab, self.xe_tab)
        """
        return jnp.where(
            lna < self.lna_xe_tab.arr[0],
            self.xe_tab.arr[0],
            jnp.where(
                lna >= self.lna_xe_tab.lastval,
                self.xe_tab.lastval,
                tools.fast_interp(lna, self.lna_xe_tab.arr[0],
                self.lna_xe_tab.arr[0] + len(self.lna_xe_tab.arr) * (self.lna_xe_tab.arr[1]-self.lna_xe_tab.arr[0]),
                self.xe_tab.arr)
            )
        )

    def _Tm_early_approx(self, lna, params):
        """
        Compute matter temperature using post-equilibrium approximation.

        Uses approximation Tm = TCMB * (1 - H/GammaCompton) for early times
        before detailed recombination calculation begins.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Matter temperature (units: eV)
        """
        TCMB = self.TCMB(lna, params)
        xe   = self.xe(lna)
        return TCMB * (1.-self.H(lna,params)/recomb_functions.Gamma_compton(xe, TCMB, params['YHe']))

    def Tm(self, lna, params):
        """
        Compute matter temperature.

        Interpolates from pre-tabulated recombination history with
        early-time approximation and late-time boundary conditions.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Matter temperature (units: eV)
        """
        return jnp.where(
            lna < self.lna_Tm_tab.arr[0],
            self._Tm_early_approx(lna, params),
            jnp.where(
                lna >= self.lna_Tm_tab.lastval,
                self.Tm_tab.lastval,
                tools.fast_interp(lna, self.lna_Tm_tab.arr[0],
                self.lna_Tm_tab.arr[0] + len(self.lna_Tm_tab.arr) * (self.lna_Tm_tab.arr[1]-self.lna_Tm_tab.arr[0]),
                self.Tm_tab.arr)
            )
        )

    def nH(self, lna, params):
        """
        Compute hydrogen number density.

        Calculates total hydrogen number density at given redshift.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Hydrogen number density (units: cm^{-3})
        """
        return (1-params['YHe']) * 3. * params['omega_b'] * cnst.H0_over_h**2 / 8 / jnp.pi / cnst.G / cnst.mH / jnp.exp(lna)**3

    def TCMB(self,lna, params):
        """
        Compute CMB temperature.

        Calculates CMB temperature at given redshift using T ∝ 1/a scaling.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            CMB temperature (units: eV)
        """
        return params['TCMB0'] / jnp.exp(lna)

    def tau_c(self, lna, params):
        """
        Compute Thomson scattering time.

        Calculates Thomson scattering time scale τc = 1/(a × ne × σT).

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Thomson scattering time (units: Mpc)
        """
        a = jnp.exp(lna)
        nH = self.nH(lna, params)
        ne = nH * self.xe(lna)
        return 1./a/ne/cnst.thomson_xsec/cnst.c*cnst.c_Mpc_over_s

    def _tabulate_optical_depth(self, params):
        """
        Tabulate optical depth from given scale factor to today.

        Integrates dκ/d(ln a) = -1/(τc × aH) backwards from today
        to compute optical depth κ(a) = ∫[a to 1] dκ/da' da'.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        array
            Tabulated optical depth values (units: dimensionless)

        Notes:
        ------
        Also computes time derivative of optical depth, which is the
        integrand involving the free electron fraction.
        """
        integrand = lambda lna, y, args: -1./self.tau_c(lna, params)/self.aH(lna, params)
        term = ODETerm(integrand)
        stepsize_controller = PIDController(pcoeff=0.4, icoeff=0.3, dcoeff=0, rtol=1.e-10, atol=1.e-10)
        adjoint=self.adjoint()
        sol = diffeqsolve(
            term,
            solver=Kvaerno5(),           
            stepsize_controller=stepsize_controller,
            t0=0.,                
            t1=-10.,                  
            dt0=-1.e-3,                 
            max_steps=2048,
            y0=0.0,                     
            saveat=SaveAt(dense=True),
            adjoint=adjoint
        )
        return sol

    def expmkappa(self, lna):
        """
        Compute optical depth.

        Interpolates from pre-tabulated optical depth history.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor

        Returns:
        --------
        float
            Optical depth (units: dimensionless)
        """

        return jnp.where(
            lna < -10.,
            0.,
            jnp.exp(-self.kappa_func.evaluate(lna))
        )

    def visibility(self, lna, params):
        """
        Compute visibility function.

        Calculates visibility function g(x) = -aH(x) × κ'(x) × exp(-κ(x))
        where ' = d/dx and x = ln a. Represents probability that a CMB
        photon observed today was last scattered at time x.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Visibility function (units: Mpc^{-1})

        Notes:
        ------
        Used in computing source functions for CMB anisotropies.
        """
        return self.expmkappa(lna)/self.tau_c(lna, params)

    ###########################################
    ### tools for computing decoupling time ###
    ###########################################

    def find_z_at_kappad_equals_one(self, z, kappa_d):
        """
        Find redshift where baryon optical depth equals unity.

        Interpolates to find z_d such that κ_d(z_d) = 1, marking
        the approximate time of baryon decoupling.

        Parameters:
        -----------
        z : array
            Redshift array
        kappa_d : array
            Baryon optical depth array

        Returns:
        --------
        float
            Decoupling redshift (units: dimensionless)
        """
        # ensure sorted ascending
        idx = jnp.argsort(z)
        z_sorted = z[idx]
        kappa_d_sorted = jnp.abs(kappa_d)[idx]

        # interpolate
        z_d = jnp.interp(1.0, kappa_d_sorted, z_sorted)
        return z_d

    def interp_rs_at_z(self, z_bg, r_s, z_d):
        """
        Interpolate sound horizon at decoupling redshift.

        Parameters:
        -----------
        z_bg : array
            Background redshift array
        r_s : array
            Sound horizon array
        z_d : float
            Decoupling redshift

        Returns:
        --------
        float
            Sound horizon at decoupling (units: Mpc)
        """
        idx = jnp.argsort(z_bg)
        z_sorted = z_bg[idx]
        rs_sorted = r_s[idx]
        return jnp.interp(z_d, z_sorted, rs_sorted)

    def R_ratio_lna(self, lna, params):
        """
        Compute baryon drag ratio.

        Calculates R = 3ρ_b/(4ρ_γ), the ratio of baryon to photon
        energy densities that appears in baryon drag calculations.

        Parameters:
        -----------
        lna : float
            Logarithm of scale factor
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Baryon drag ratio (units: dimensionless)
        """

        rho_b = 0.
        rho_g = 0.

        for s in self.species_list:
            if s.name == "Photon":
                rho_g += s.rho(lna, params)
            elif s.name == "Baryon":
                rho_b += s.rho(lna, params)

        return 3. * rho_b / (4 * rho_g)

    @jax.named_scope("tabulate kappa d")
    def _tabulate_kappa_d(self, params):
        """
        Tabulate baryon optical depth.

        Integrates dκ_d/d(ln a) = -1/(τc × aH × R) backwards from today
        to compute baryon optical depth including drag effects.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        array
            Tabulated baryon optical depth values (units: dimensionless)
        """
        integrand = lambda lna, y, args: jnp.float64(-1./self.tau_c(lna, params)/self.aH(lna, params)/(self.R_ratio_lna(lna, params)))
        term = ODETerm(integrand)
        stepsize_controller = PIDController(pcoeff=0.4, icoeff=0.3, dcoeff=0, rtol=1.e-3, atol=1.e-6)
        adjoint=self.adjoint()
        
        solution = diffeqsolve(
            term,
            solver=Tsit5(),            # Kvaerno5 is just slower but gives same result
            stepsize_controller=stepsize_controller,
            t0=self.lna_tau_tab[-1],                 # Initial x value (~0 in this case)
            t1=self.lna_tau_tab[0],                  # Final x value (smallest x value)
            dt0=-1e-3,                  # Initial step size
            max_steps=2048,
            y0=0.0,                     # Initial value tau(x=0) = 0
            saveat=SaveAt(ts=self.lna_tau_tab[::-1]), # Save at all points in x, reverse order since integrating backwards
            adjoint=adjoint
        )
        result = solution.ys[::-1]
        return result

    @jax.named_scope("tabulate rs")
    def _tabulate_rs(self, params):
        """
        Tabulate sound horizon evolution.

        Integrates drs/d(ln a) = cs/aH from early times to today
        where cs = 1/√(3(1+R)) accounts for baryon loading.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        array
            Tabulated sound horizon values (units: Mpc)
        """
         # initial condition assuming cs**2 = 1/3 at early times
        rs0 = 1./jnp.sqrt(3) / (self.aH( self.lna_tau_tab[0], params ))

        integrand = lambda lna, y, args: 1./jnp.sqrt(3*(1+self.R_ratio_lna(lna, params))) / (self.aH(lna, params))
        term = ODETerm(integrand)
        stepsize_controller = PIDController(pcoeff=0.4, icoeff=0.3, dcoeff=0, rtol=1.e-3, atol=1.e-6)
        adjoint=self.adjoint()
        
        solution = diffeqsolve(
            term,
            solver=Tsit5(),
            stepsize_controller=stepsize_controller,
            t0=self.lna_tau_tab[0],                 # reversed direction since I know rs at early times
            t1=self.lna_tau_tab[-1],
            dt0=1e-3,
            max_steps=2048,
            y0=rs0,
            saveat=SaveAt(ts=self.lna_tau_tab),
            adjoint=adjoint
        )
        result = solution.ys
        return result
        

    def z_d(self, params):
        """
        Compute baryon decoupling redshift.

        Finds redshift where κ_d = 1 as estimate of when baryons
        decouple from photons.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Decoupling redshift (units: dimensionless)
        """
        return self.find_z_at_kappad_equals_one(1/jnp.exp(self.lna_tau_tab) - 1, self._tabulate_kappa_d(params))

    def rs_d(self, params):
        """
        Compute sound horizon at decoupling.

        Finds value of sound horizon at baryon decoupling redshift z_d.

        Parameters:
        -----------
        params : dict
            Cosmological parameters

        Returns:
        --------
        float
            Sound horizon at decoupling (units: Mpc)
        """
        return self.interp_rs_at_z(1/jnp.exp(self.lna_tau_tab) - 1, self._tabulate_rs(params), self.z_d(params))

class ReionizationModel(eqx.Module):
    """
    Object for computing the reionization correction to the free electron fraction.
    Provides the base methods 

    xe_reion : calculates the tanh electron fraction correction at redshifts lna, given z_reion and params
    tau_reion_fn : calculates the optical depth to reionization.
    
    At the moment we only support the CAMB tanh parameterization, but we need different approaches
    based on whether the use inputs the optical depth tau_reion or the reionization redshift z_reion.

    """

    z_reion : jnp.float64
    tau_reion : jnp.float64

    def xe_reion(self, lna, z_reion, params):
        """
        Passing in an lna array should get you the correct tanh patching based on the
        reionization parameter. 
        """
        fHe = params['YHe'] / 4 / (1-params['YHe'])
        z = 1/jnp.exp(lna) - 1
        y = (1+z)**(params["exp_reion"])

        y_reion = (1+z_reion)**(params["exp_reion"])
        Delta_y_reion = params["exp_reion"] * (1+z_reion)**(params["exp_reion"]-1) * params["Delta_z_reion"]
        tanh_arg = (y_reion - y) / Delta_y_reion
        xe_reion_H = (1+fHe)/2 * (1 + jnp.tanh(tanh_arg))

        # The above accounts for hydrogen and the first ionization level of helium.
        # Let's also account for the second ionization of helium:
        tanh_arg_He = (params["z_reion_He"] - z)/params["Delta_z_reion_He"]
        xe_reion_HeII = fHe/2 * (1 + jnp.tanh(tanh_arg_He))

        return xe_reion_H + xe_reion_HeII

    def tau_reion_fn(self, z_reion, BG, params):
        lna_axis = jnp.linspace(-5., 0., 2000)
        xe_reion_correction = self.xe_reion(lna_axis, z_reion, params)
        # Free electron number density belonging only to reionized hydrogen. 
        ne = BG.nH(lna_axis, params) * xe_reion_correction
        Gamma = jnp.exp(lna_axis)*ne*cnst.thomson_xsec*cnst.c/cnst.c_Mpc_over_s
        aH = BG.aH(lna_axis, params)
        # Optical depth integrand
        integrand = Gamma/aH
        return jnp.trapezoid(integrand, lna_axis)

class ReionizationModelFromZ(ReionizationModel):
    """
    Concrete extension of the base ReionizationModel Class.
    This object is used when the user direcly inputs the redshift of reionization.
    In this case the tanh correction and the optical depth can be computed directly,
    and simply returned.
    """

    def __init__(self, BG, params):
        self.z_reion = params.get("z_reion", jnp.array(7.6711))
        self.tau_reion = self.tau_reion_fn(self.z_reion, BG, params)

class ReionizationModelFromTau(ReionizationModel):

    """
    Concrete extension of the base ReionizationModel Class.
    This object is used when the user inputs the optical depth and wishes to infer the redshift.
    The init finder will use an optimistix root finder to find the appropriate redshift.
    Then the appropriate tanh correction may be called and returned, as well as the inferred reionization redshift.
    """

    def __init__(self, BG, params):

        def tau_target_fn(z_reion, args):
            target = args
            return self.tau_reion_fn(z_reion, BG, params) - target

        solver = optx.Newton(rtol=1e-5, atol=1e-5)
        sol = optx.root_find(tau_target_fn, solver, 7.6, params.get("tau_reion", jnp.array(0.05430842)))
        self.z_reion = sol.value
        self.tau_reion = params.get("tau_reion", jnp.array(0.05430842))


from functools import partial

import jax
from jax import lax
from jax import numpy as jnp 
from jax.scipy.special import zeta, gammaln, i0, i1, bernoulli
from jax.scipy.special import gamma as jax_gamma

euler_gamma = 0.57721566490153286061
zeta_3 = 1.202056903159594

# Number of terms for Li series. 
L = 60

# List of Bernoulli numbers B_n 
bernoulli_ary = bernoulli(L) 

def comb(N,k):
    """
    Combinatoric factor N! / (k! (N-k)!). 
    """
    return jnp.exp((gammaln(N+1) - gammaln(k+1) - gammaln(N-k+1) ))

def Bernoulli(n, x): 
    """
    The Bernoulli polynomial B_n(x), n < 60. See Wikipedia article. 
    """

    return lax.fori_loop(
        0, n+1, lambda i, val: val+bernoulli_ary[i] * comb(n,i) * x**(n-i), 0
    )

def gamma(x): 
    """
    Gamma function using the Lanczos approximation. Needed for imaginary
    arguments. 
    """
    def lanczos(x): 

        g = 7

        p_vals = jnp.array([
            0.99999999999980993,
            676.5203681218851,
            -1259.1392167224028,
            771.32342877765313,
            -176.61502916214059,
            12.507343278686905,
            -0.13857109526572012,
            9.9843695780195716e-6,
            1.5056327351493116e-7
        ])

        eps = 1e-06

        z = x
        z_fill = z - 1
        denom = z_fill + jnp.arange(1, len(p_vals))
        x_fill = p_vals[0] + jnp.sum(p_vals[1:] / denom)
        t = z_fill + g + 0.5

        result1 = (
            jnp.sqrt(2 * jnp.pi) * t ** (z_fill + 0.5) * jnp.exp(-t) * x_fill
        )
        result2 = jnp.where(
            jnp.abs(result1.imag) <= eps, result1.real, result1
        )
        return result2

    
    def reflect(x): 

        return jnp.pi / jnp.sin(jnp.pi * x) * lanczos(1. - x) 

    return lax.cond(x < 0.5, reflect, lanczos, x)


def Riemann_zeta(n): 
    """
    Riemann zeta function. Works for n > -60. Fixes results
    for negative arguments. 
    """

    return jnp.where(
        n > 0, zeta(n, 1), jnp.where(
            (n < 0) & (-n % 2 == 0), 0., 
            (-1.)**(-n) * bernoulli_ary[-n+1] / (-n + 1)
        )
    )

@partial(jax.jit, static_argnums=(0,))
def Li(n, z): 
    """
    Polylogarithm of order n and argument z. 
    """
    def _Li_z_small(z): 
    
        return lax.fori_loop(1, L, lambda j,val: val + z**j / j**n, 0)

    def _Li_z_intermed(z): 

        # Oddly enough, the fastest way to do this. 
        zeta_ary = jnp.concatenate((
            jnp.array([Riemann_zeta(n - m) for m in jnp.arange(n-1)]), 
            jnp.array([0.]), 
            jnp.array([Riemann_zeta(n - m) for m in jnp.arange(n, L)])
        ))

        zeta_series_term = jnp.sum(
            zeta_ary * jnp.concatenate(
                (jnp.array([1., jnp.log(z+0j)]), 
                jnp.log(z+0j)**jnp.arange(2, L))
            ) / jax_gamma(jnp.arange(L) + 1.)
        )

        H_n_m_1 = jnp.sum(1. / jnp.arange(1, n))
        harmonic_term = jnp.where(
            jnp.isclose(z - 1., 0), 0., 
            jnp.log(z+0j)**(n-1) / jax_gamma(n) * (
                H_n_m_1 - jnp.log(-jnp.log(
                    jnp.where(jnp.isclose(z - 1., 0), 2., z)+0j) + 0j
                )
            )
        )

        res = zeta_series_term + harmonic_term 

        return jnp.real(res)
    
    def _Li_z_large(z):

        recip_Li = lax.fori_loop(1, L, lambda j,val: val + (1/z)**j / j**n, 0)

        B_n = Bernoulli(n, jnp.log(z+0j)/(2 * jnp.pi * 1j))
 
        return jnp.real(
            - (-1)**n * recip_Li 
            - (2*jnp.pi*1j)**n / jax_gamma(n + 1) * B_n 
        )

    small_range = jnp.abs(z) <= 0.5

    intermed_range = (0.5 < jnp.abs(z)) & (jnp.abs(z) < 2)

    large_range = jnp.abs(z) > 2

    return jnp.where(
        small_range, _Li_z_small(jnp.where(small_range, z, 3.)), jnp.where(
            intermed_range, _Li_z_intermed(
                jnp.where(intermed_range, z, 3.)
            ), jnp.where(
                large_range, _Li_z_large(jnp.where(large_range, z, 3.)), 0.
            )
        ) 
    )

def K0(z): 
    """
    Modified Bessel function of the second kind of order 0. 
    See Zhang and Jin for algorithm. 
    """

    def K0_small(z):

        # n = 30 sufficient for abstol ~ 1e-12 and reltol ~ 1e-8
        int_ary = jnp.arange(1., 31)

        harmonic_ary = jnp.cumsum(1. / jnp.arange(1, 31))

        return -(jnp.log(z/2.) + euler_gamma) * i0(jnp.where(z < 600., z, 600.)) + jnp.sum(
            harmonic_ary * (z/2.)**(2.*int_ary) / jax_gamma(int_ary+1)**2.
        )

    def K0_large(z): 

        # n = 10 sufficient for abstol ~ 1e-12 and reltol ~ 1e-8

        int_ary = jnp.arange(1., 11)

        prod_term = jnp.cumprod(-(2.*int_ary - 1.) / (2.*int_ary) * (2.*int_ary - 1.)**2.) 

        res = 1. / 2. / z / i0(jnp.where(z < 600., z, 600.)) * (1. + jnp.sum((-1)**int_ary * prod_term / (2.*z)**(2.*int_ary)))

        return jnp.where(z < 600, res, 0.)

    return lax.cond(z < 9, K0_small, K0_large, z)

def K1(z): 
    """
    Modified Bessel function of the second kind of order 1. 
    """

    def K1_small(z): 

        return (1 / z - i1(z) * K0(z)) / i0(z)

    return jnp.where(z < 600., K1_small(jnp.where(z < 600. , z, 600.)), 0.)


def K2(z): 
    """
    Modified Bessel function of the second kind of order 2. 
    """

    return K0(z) + 2 / z * K1(z) 




import sys
sys.path.append('..')

import jax.numpy as jnp
import equinox as eqx
from diffrax import diffeqsolve, ODETerm, Tsit5, Kvaerno3, PIDController, SaveAt, ForwardMode

from . import nuclear as nucl
from . import const as const 
from .const import ma, me, mn, mp
from . import weak_rates as wr
from . import thermo as thermo
from .thermo import rho_EM_std_v, p_EM_std_v, nB
from .special_funcs import zeta_3 

class AbundanceModel(eqx.Module): 
    """
    Abundance model and BBN abundance prediction. 

    Attributes
    ----------
    nuclear_net : NuclearRates
        Nuclear network to be used for BBN prediction. 
    weak_rates : WeakRates
        Weak rates for neutron-proton interconversion. 
    species_dict : dict
        Dictionary of species considered in LINX. 
    species_Z : list
        Number of protons in each species. 
    species_N : list
        Number of neutrons in each species. 
    species_A : list
        Atomic mass number of each species. 
    species_excess_mass : list
        Excess mass (mass - A*amu) of each species. 
    species_spin : list
        Spin of each species. 
    species_binding_energy : list
        Binding energy of each species. 
    species_mass : list
        Mass of each species. 
    """
    
    nuclear_net : nucl.NuclearRates  
    weak_rates : wr.WeakRates 
    species_dict : dict
    species_Z : list
    species_N : list
    species_A : list 
    species_excess_mass : dict
    species_spin : list
    species_binding_energy : list
    species_mass : list

    adjoint : "diffrax.adjoint" = eqx.field(static=True)

    def __init__(self, nuclear_net, weak_rates=wr.WeakRates(), adjoint=ForwardMode):

        self.adjoint = adjoint
        self.nuclear_net = nuclear_net
        self.weak_rates = weak_rates

        self.species_dict = {
            0:'n', 1:'p', 2:'d', 3:'t', 4:'He3', 5:'a', 6:'Li7', 7:'Be7', 
            8: 'Li6', 9: 'He6', 10: 'Li8', 11:'B8'
        }

        self.species_Z = jnp.array([0, 1, 1, 1, 2, 2, 3, 4, 3, 2, 3, 5])

        self.species_N = jnp.array([1, 0, 1, 2, 1, 2, 4, 3, 3, 4, 5, 3])

        self.species_A = self.species_Z + self.species_N 

        # in MeV
        self.species_excess_mass = jnp.array([
            8071.3171, 7288.9706, 13135.722, 14949.81, 
            14931.218, 2424.9156, 14907.105, 15769.,
            14086.8789, 17592.10, 20945.80, 22921.6
        ]) * 1e-3

        self.species_spin = jnp.array([
            1./2., 1./2., 1., 1./2., 1./2., 0.,
            3./2., 3./2., 1., 0., 2., 2.
        ])

        # in MeV
        self.species_binding_energy = (
            self.species_N * self.species_excess_mass[0]
            + self.species_Z * self.species_excess_mass[1]
            - self.species_excess_mass 
        )

        # in MeV
        self.species_mass = (
            self.species_A * ma + self.species_excess_mass - self.species_Z * me
        )

    @eqx.filter_jit
    def __call__(
        self, rho_g_vec, rho_nu_vec, rho_NP_vec, P_NP_vec, 
        a_vec=None, t_vec=None, 
        eta_fac=jnp.asarray(1.), tau_n_fac = jnp.asarray(1.), 
        nuclear_rates_q=None, 
        Y_i=None, T_start=None, T_end=None, sampling_nTOp=150, 
        rtol=1e-6, atol=1e-9, solver=Kvaerno3(),
        max_steps=4096,
        save_history=False
    ):
        """
        Calculate BBN abundance. 

        Parameters
        ----------
        rho_g_vec : array
            Energy density of photons in MeV^4. 
        rho_nu_vec : array
            Energy density of a single neutrino species in MeV^4 
            (all neutrinos assumed to have the same temperature). 
        rho_NP_vec : array
            Energy density of all new physics particles in MeV^4. 
        P_NP_vec : array
            Pressure of all new physics particles in MeV^4. 
        a_vec : array, optional
            Scale factor. If `None`, will be computed in function. 
        t_vec : array, optional
            Time in seconds. If `None`, will be computed in function. 
        eta_fac : float, optional
            Rescaling factor for baryon-to-photon ratio, 1 for fiducial value 
            in `const.eta0` (or `const.Omegabh2`). 
        tau_n_fac : float, optional
            Rescaling factor for neutron decay lifetime, 1 for fiducial value 
            in `const.eta0` (or `const.Omegabh2`). 
        nuclear_rates_q : array, optional
            q ~ N(0,1) specifies the nuclear rate in its log-normal 
            distribution. If not specified, will be taken to be `q = 0`. 
        Y_i : tuple of float, optional
            Initial abundances :math:`n_i/n_b` for species. Length must be equal to 
            `self.nuclear_net.max_i_species`. Must specify `T_start` and `T_end` if not `None`. 
        T_start : float
            Temperature in MeV to start integration. Must specify `Y_i` and `T_end` if not `None`, otherwise `const.T_start` used. 
        T_end : float
            Temperature in MeV to end integration. 
        sampling_nTOp : int
            Number of points to subdivide (`T_end`, `T_start`) for neutron-proton interconversion rate interpolation table. 
        rtol : float, optional
            Relative tolerance of the abundance solver. Default is `1e-4`. 
        atol : float, optional
            Absolute tolerance of the abundance solver. Default is `1e-9`. 
        max_steps : int, optional
            Maximum number of steps taken by the solver. Default is `4096`. 
            Increasing this slows down the code, while decreasing this could 
            mean that the solver cannot complete the solution. 
        solver : Diffrax ODE solver 
            The Diffrax ODE solver to use. A stiff solver is recommended. 
            Default is the 3rd order Kvaerno solver. 
        save_history : bool
            If `True`, full solution is returned with temperature and time 
            abscissa.

        Returns
        -------
        tuple of array or array
            If `save_history` is set to `True`, a tuple containing an array of EM temperatures, an array of times, and a Diffrax `Solution` instance, which can be called as a function of time. Otherwise, returns yields of all species considered in `self.nuclear_net`.  

        """

        print('Compiling abundance model...')

        if Y_i is not None: 
            if T_start is None: 
                raise TypeError('Specifying Y_i requires specifying a T_start')
        if T_start is not None: 
            if Y_i is None: 
                raise TypeError('Specifying T_start requires specifying Y_i')

        if nuclear_rates_q is None: 

            nuclear_rates_q = jnp.array(
                [0. for _ in self.nuclear_net.reactions]
            )

        if t_vec is None: 
            t_vec = self.get_t(rho_g_vec, rho_nu_vec, rho_NP_vec, P_NP_vec)

        if a_vec is None: 
            a_vec = self.get_a(rho_g_vec, rho_nu_vec, rho_NP_vec, P_NP_vec)

        if T_start is None: 

            T_start  = const.T_start

        if T_end is None: 

            T_end  = const.T_end

        # These are in MeV
        T_g_vec  = thermo.T_g(rho_g_vec)
        T_nu_vec = thermo.T_nu(rho_nu_vec) 

        a_start  = jnp.exp(
            jnp.interp(
                jnp.log(T_start), 
                jnp.flip(jnp.log(T_g_vec)), 
                jnp.flip(jnp.log(a_vec)), 
                left=jnp.log(a_vec[-1]), right=jnp.log(a_vec[0])
            )
        )

        t_start = jnp.exp(
            jnp.interp(
                jnp.log(T_start), 
                jnp.flip(jnp.log(T_g_vec)), 
                jnp.flip(jnp.log(t_vec)),
                left=jnp.log(t_vec[-1]), right=jnp.log(t_vec[0])
            )
        )

        t_end = jnp.exp(
            jnp.interp(
                jnp.log(T_end), 
                jnp.flip(jnp.log(T_g_vec)), 
                jnp.flip(jnp.log(t_vec)),
                left=jnp.log(t_vec[-1]), right=jnp.log(t_vec[0])
            )
        )

        ##################################
        # Weak Rates                     #
        ##################################

        T_interval_nTOp, nTOp_frwrd, nTOp_bkwrd = self.weak_rates(
            jnp.array([T_g_vec, T_nu_vec]), 
            T_start=T_start, T_end=T_end, sampling_nTOp=sampling_nTOp
        )

        ##################################
        # Initialization of Abundances   #
        ##################################
        
        
        if Y_i is None: 
            # If not provided, initialized to const.T_start. 

            # Neutron and proton yields, based on rates. 
            Yn_i = nTOp_bkwrd[0] / (nTOp_frwrd[0] + nTOp_bkwrd[0])
            Yp_i = 1. - Yn_i 

            # Other elements start at statistical equilibrium. 
            n_CMB_start = thermo.n_massless_BE(T_start, 0., 2.)
            eta_T_start = nB(a_start, eta_fac=eta_fac) / n_CMB_start

            Y_YNSE = self.YNSE(Yn_i, Yp_i, const.T_start, eta_T_start) 

            Y_others_i = Y_YNSE[2:self.nuclear_net.max_i_species]

            Y_i = (Yn_i, Yp_i) + tuple(Y_others_i)

        if save_history: 
            saveat = SaveAt(dense=True) 
        else: 
            # Default SaveAt
            saveat = SaveAt(t1=True)

        sol = diffeqsolve(
            ODETerm(self.Y_prime), solver, 
            t0=t_start, t1=t_end, dt0=None, y0=Y_i, 
            args = (
                a_vec, t_vec, T_g_vec, T_interval_nTOp, nTOp_frwrd, 
                nTOp_bkwrd, eta_fac, tau_n_fac, nuclear_rates_q
            ), saveat=saveat, stepsize_controller = PIDController(
                rtol=rtol, atol=atol,
            ),
            max_steps=max_steps,
            adjoint = self.adjoint()
        )

        if save_history: 
            return sol 
        else: 
            Y_f = jnp.array(sol.ys).flatten()
            return Y_f

    @eqx.filter_jit
    def get_t(self, rho_g_vec, rho_nu_vec, rho_NP_vec, P_NP_vec):
        """ 
        Time elapsed. 

        Parameters
        ----------
        rho_g_vec : array
            Energy density of photons. 
        rho_nu_vec : array
            Energy density of one species of neutrinos (assumed identical for all species). 
        rho_NP_vec : array
            Energy density of all new physics fluids. 
        P_NP_vec : array
            Pressure of all new physics fluids. 

        Returns
        -------
        array
            Array of times in seconds corresponding to physical parameters 
            above. Initial time taken to be 1 / (2H). 
        
        """

        T_g_vec  = thermo.T_g(rho_g_vec)

        rho_tot_vec = (
            rho_EM_std_v(T_g_vec) + 3 * rho_nu_vec + rho_NP_vec 
        )

        P_tot_vec = p_EM_std_v(T_g_vec) + 3 * (rho_nu_vec/3) + P_NP_vec

        def P_tot(rho_tot): 

            return jnp.exp(
                jnp.interp(
                    jnp.log(rho_tot), 
                    jnp.flip(jnp.log(rho_tot_vec)), 
                    jnp.flip(jnp.log(P_tot_vec))
                )
            )

        def dt_prime(rho_tot, t, args): 

            return 1. / (
                -3. * thermo.Hubble(rho_tot) *(rho_tot + P_tot(rho_tot))
            )

        rho_tot_init = rho_tot_vec[0]
        rho_tot_fin  = rho_tot_vec[-1]

        sol_t = diffeqsolve(
            ODETerm(dt_prime), Tsit5(),
            t0=rho_tot_init, t1=rho_tot_fin,
            y0=1. / (2 * thermo.Hubble(rho_tot_init)),
            dt0=None, max_steps=4096,
            saveat=SaveAt(ts=rho_tot_vec),
            stepsize_controller=PIDController(rtol=1e-8, atol=1e-10),
            adjoint = self.adjoint()
        )

        return sol_t.ys

    @eqx.filter_jit
    def get_a(self, rho_g_vec, rho_nu_vec, rho_NP_vec, P_NP_vec): 
        """ 
        Scale factor. 

        Parameters
        ----------
        rho_g_vec : array
            Energy density of photons. 
        rho_nu_vec : array
            Energy density of one species of neutrinos (assumed identical for 
            all species). 
        rho_NP_vec : array
            Energy density of all new physics fluids. 
        P_NP_vec : array
            Pressure of all new physics fluids. 

        Returns
        -------
        array
            Array of scale factors corresponding to physical parameters above. 
            
        Notes
        -----
        The final entry `a[-1]` is given by `const.T0CMB / T_gamma[-1]`, where `const.T0CMB` is the CMB temperature measured today, and `T_gamma[-1]` is the temperature of photons in the last entry of `rho_g_vec`. In other words, we assume no subsequent entropy dump in the electromagnetic sector.
        
        """

        T_g_vec  = thermo.T_g(rho_g_vec)

        rho_tot_vec = (
            rho_EM_std_v(T_g_vec) + 3 * rho_nu_vec + rho_NP_vec 
        )  

        P_tot_vec = p_EM_std_v(T_g_vec) + 3 * (rho_nu_vec/3) + P_NP_vec

        def P_tot(rho_tot): 

            return jnp.exp(
                jnp.interp(
                    jnp.log(rho_tot), jnp.flip(jnp.log(rho_tot_vec)), 
                    jnp.flip(jnp.log(P_tot_vec))
                )
            )   

        def dlna_prime(rho_tot, t, args): 

            return 1. / (-3. * (rho_tot + P_tot(rho_tot)))
        
        rho_tot_init = rho_tot_vec[0]
        rho_tot_fin  = rho_tot_vec[-1]

        # a_0 = 1 arbitrarily, will rescale later. 
        sol_lna = diffeqsolve(
            ODETerm(dlna_prime), Tsit5(),
            t0=rho_tot_init, t1=rho_tot_fin,
            y0=0., dt0=None, max_steps=4096,
            saveat=SaveAt(ts=rho_tot_vec),
            stepsize_controller=PIDController(rtol=1e-8, atol=1e-10),
            adjoint = self.adjoint()
        )

        a_fin = const.T0CMB / T_g_vec[-1] 

        a_vec = jnp.exp(sol_lna.ys)
        # Rescale so that the last a is given by T_g_vec[-1] / TCMB today. 
        a_vec = a_vec / a_vec[-1] * a_fin

        return a_vec


    @eqx.filter_jit
    def Y_prime(self, t, Y, args):
        """Returns :math:`dY_i/dt` for this abundance model. 

        Parameters
        ----------
        t : float 
            Time at which to evaluate :math:`dY_i/dt`.
        Y : array
            Array of abundances for evaluating :math:`dY_i/dt`. 
        args : tuple of arrays
            Other relevant information for evaluating the derivative. These are respectively, 0) an array of scale factors; 1) an array of times; 2) an array of EM sector temperatures; 3) an array representing the abscissa of EM sector temperatures for evaluating weak rates; 4) an array of n -> p rates to interpolate over; 5) an array of p -> n rates to interpolate over; 6) the rescaling factor for baryon-to-photon ratio `eta_fac`; 7) the rescaling factor for neutron decay lifetime `tau_n_fac` and 8) the array rescaling nuclear rates, `nuclear_rates_q`. 

        Returns
        -------
        array
            :math:`dY_i/dt` at the given time and at the present abundance levels.
        
        """

        a_vec_in    = args[0]
        t_vec_in    = args[1]
        T_g_vec_in  = args[2]
        T_interval_in = args[3]
        nTOp_frwrd_vec_in = args[4]
        nTOp_bkwrd_vec_in = args[5]
        eta_fac = args[6]
        tau_n_fac = args[7] 
        nuclear_rates_q = args[8]
        
        a_in  = a_vec_in[0]
        a_fin = a_vec_in[-1]

        a = jnp.interp(
            t, t_vec_in, a_vec_in,
            left=a_in,right=a_fin
        )
        
        # Baryon density rescaled by eta_fac. 
        n0B = eta_fac*const.n0CMB*const.eta0
        rho0BmaOvermB = ma * n0B
        # number density times amu. 
        rhoBBN = rho0BmaOvermB * const.MeV4_to_gcmm3/a**3 

        T_t = jnp.interp(
            t, t_vec_in, T_g_vec_in, left=T_g_vec_in[0],right=T_g_vec_in[-1]
        )

        dY = self.nuclear_net(
            Y, T_t, rhoBBN, T_interval_in, nTOp_frwrd_vec_in,
            nTOp_bkwrd_vec_in, tau_n_fac=tau_n_fac, 
            nuclear_rates_q=nuclear_rates_q
        )

        return dY

    def YNSE(self, Yn, Yp, T, eta): 
        """
        Nuclear statistical equilibrium yields for all species. 

        Parameters
        ----------
        Yn : float
            The yield :math:`n_n / n_b` of free neutrons. 
        Yp : float
            The yield :math:`n_p / n_b` of free protons. 
        T : float
            The temperature of the baryons in MeV.
        eta : float
            The baryon-to-photon ratio. 

        Returns
        -------
        array
            Yields for all species considered in LINX (13 of them). 
        """

        A32Overmn = (
            self.species_mass / (
                mn**(self.species_A - self.species_Z) 
                * mp**self.species_Z
            )
        )**(3/2)

        return (
            (2 * self.species_spin + 1) * zeta_3**(self.species_A-1) 
            * jnp.pi**((1-self.species_A)/2) * 2**((3*self.species_A-5)/2) 
            * A32Overmn * T**(3/2*(self.species_A-1)) 
            * eta**(self.species_A-1) * Yp**self.species_Z 
            * Yn**(self.species_A-self.species_Z) 
            * jnp.exp(self.species_binding_energy / T)
        )

